{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "(foundations:symbolic_interpreter)=\n",
    "# A Symbolic Interpreter and Correctness of Symbolic Transitions"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Executing a program with the concrete minipy interpreter updates the passed environment, and potentially raises an exception. Symbolic execution, in contrast, takes into account all feasible paths through a program. We distinguish between *online* and *offline* SE (see, e.g., {cite}`baldoni.coppa.ea-18`). Online SE executes multiple symbolic paths simultaneously: For example, executing an `if` statement leads to two paths which are both explored. Offline SE reasons about one single path at a time. *Concolic execution*, where SE is guided by a single *concrete* input, is an example of the latter class. In this section, we regard *online* SE. Symbolically executing a program in an online symbolic interpreter results in multiple output environments. Also, such an interpreter cannot simply raise an exception, as this would result in a premature termination of the analysis, skipping other feasible program paths. The consequences for our static, online symbolic interpreter are:\n",
    "\n",
    "* Instead of updating the initial symbolic environment, the interpreter produces potentially multiple new environments. Our interpreter even outputs all *intermediate* states, such that the output is a *Symbolic Execution Tree* (SET).\n",
    "* Thrown exceptions are stored in the symbolic environment instead of being raise. Thus, we need another field in our `SymbolicEnvironment` class.\n",
    "\n",
    "We implement these changes subsequently. Afterward, we show how to implement a static, online symbolic interpreter for minipy, and introduce the correctness properties *exhaustiveness* and *precision* for symbolic execution. Finally, we show how these definitions can be used to *automatically test* the symbolic interpreter. Using this methodology, we discovered two bugs of precious versions of our implementation, one of which is related to a subtle difference in the division semantics in z3 and Python."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Symbolic Execution Trees\n",
    "\n",
    "We first update the implementation of `SymbolicEnvironment` to include abrupt completion information:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "import utils\n",
    "from minipy import *\n",
    "from constraint_solving import *\n",
    "from semantics import *"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SymbolicEnvironment(SymbolicEnvironment):\n",
    "    def __init__(self,\n",
    "                 store: Optional[SymbolicStore] = None,\n",
    "                 path_constraints: Optional[Set[z3.BoolRef]] = None,\n",
    "                 abrupt_completion: Optional[Exception] = None,\n",
    "                 functions: Optional[Dict[str, Tuple[Tuple[Variable], Type, Callable]]] = None):\n",
    "        super().__init__(store, path_constraints, functions)\n",
    "        self.abrupt_completion = abrupt_completion\n",
    "\n",
    "    def copy(self):\n",
    "        return SymbolicEnvironment(self.store, self.path_constraints, self.abrupt_completion, self.functions)\n",
    "\n",
    "    def set(self, variable: Variable, value: SymbolicValueType) -> 'SymbolicEnvironment':\n",
    "        return SymbolicEnvironment(self.store.set(variable, value), self.path_constraints,\n",
    "                                   self.abrupt_completion, self.functions)\n",
    "\n",
    "    def set_abrupt_completion(self, exc: Optional[Exception]) -> 'SymbolicEnvironment':\n",
    "        return SymbolicEnvironment(self.store, self.path_constraints, exc, self.functions)\n",
    "\n",
    "    def add_constraint(self, constraint: z3.BoolRef) -> 'SymbolicEnvironment':\n",
    "        if z3.is_false(z3.simplify(constraint)):\n",
    "            new_constraints = {z3.BoolVal(False)}\n",
    "        elif z3.is_true(z3.simplify(constraint)):\n",
    "            return self\n",
    "        else:\n",
    "            old_constraints = {other_constraint for other_constraint in self.path_constraints\n",
    "                               if not implies(constraint, other_constraint)}\n",
    "            new_constraints = old_constraints | {constraint}\n",
    "\n",
    "\n",
    "        return SymbolicEnvironment(self.store, new_constraints, self.abrupt_completion, self.functions)\n",
    "\n",
    "\n",
    "    def add_function(self, name: str, params: Tuple[Variable], type: Type, impl: Callable) -> 'SymbolicEnvironment':\n",
    "        new_functions = copy.deepcopy(self.functions)\n",
    "        new_functions[name] = (params, type, impl)\n",
    "        return SymbolicEnvironment(self.store, self.path_constraints, self.abrupt_completion, new_functions)\n",
    "\n",
    "    def __str__(self):\n",
    "        return (\"(\" +\n",
    "                (\"{}\" if not self.path_constraints else str(self.path_constraints)) + \", \" +\n",
    "                str(self.store) +\n",
    "                (\"\" if self.abrupt_completion is None else repr(self.abrupt_completion)) +\n",
    "                \")\")\n",
    "\n",
    "    def to_html(self):\n",
    "        prev_html = z3.in_html_mode()\n",
    "        z3.set_html_mode(True)\n",
    "\n",
    "        result = f\"({self.path_constraint_to_html()}, {repr(self.abrupt_completion)}, {self.store.to_html()})\"\n",
    "\n",
    "        z3.set_html_mode(prev_html)\n",
    "        return result\n",
    "\n",
    "    def path_constraint_to_html(self):\n",
    "        prev_html = z3.in_html_mode()\n",
    "        z3.set_html_mode(True)\n",
    "\n",
    "        path_constraint_string = \"{\"\n",
    "        if self.path_constraints:\n",
    "            path_constraint_string += \", \".join(map(z3_html_escape, map(z3.obj_to_string, self.path_constraints)))\n",
    "        path_constraint_string += \"}\"\n",
    "        path_constraint_string = path_constraint_string.replace(\"\\n\", \"<BR/>\").replace(\" \", \"&nbsp;\")\n",
    "\n",
    "        z3.set_html_mode(prev_html)\n",
    "        return path_constraint_string\n",
    "\n",
    "\n",
    "    def __repr__(self):\n",
    "        result = f\"SymbolicEnvironment({repr(self.store)}, \" \\\n",
    "                 f\"{repr(self.path_constraints)}, \" \\\n",
    "                 f\"{repr(self.abrupt_completion)}, \"\n",
    "        result += repr({f_name: f\"fun {f_name}{params} -> \"\n",
    "                                f\"{type}\" for f_name, (params, type, _) in self.functions.items()})\n",
    "        return f\"{result})\"\n",
    "\n",
    "    def __eq__(self, other):\n",
    "        return isinstance(other, SymbolicEnvironment) and self.store == other.store and \\\n",
    "               self.path_constraints == other.path_constraints and \\\n",
    "               self.functions == other.functions and \\\n",
    "               (self.abrupt_completion == other.abrupt_completion is None or\n",
    "                ExceptionWrapper(self.abrupt_completion) == ExceptionWrapper(other.abrupt_completion))\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Next, we define a class for nodes in an SET. An SET node comprises a symbolic environment and a string representation of the next instruction to execute. We also define a `to_html` method to produce pretty output."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SETNode:\n",
    "    def __init__(self, environment: SymbolicEnvironment, pc: Optional[Union[ParseTree, str]] = None):\n",
    "        self.environment = environment\n",
    "        self.pc = pc\n",
    "\n",
    "    @property\n",
    "    def pc(self):\n",
    "        return self.__pc\n",
    "\n",
    "    @pc.setter\n",
    "    def pc(self, pc: Optional[Union[ParseTree, str]]):\n",
    "        if pc is None:\n",
    "            self.__pc = \"\"\n",
    "        else:\n",
    "            assert isinstance(pc, str)\n",
    "            self.__pc = pc.split(\"\\n\")[0].strip()\n",
    "\n",
    "    def to_html(self):\n",
    "        prev_html = z3.in_html_mode()\n",
    "        z3.set_html_mode(True)\n",
    "\n",
    "        result = f'<TABLE BORDER=\"0\"><TR>'\n",
    "        if \"<BR/>\" in self.environment.to_html():\n",
    "            result += f'<TD BALIGN=\"LEFT\" ALIGN=\"LEFT\">{self.environment.to_html()}</TD>'\n",
    "        else:\n",
    "            result += f'<TD>{self.environment.to_html()}</TD>'\n",
    "        result += \"</TR>\"\n",
    "\n",
    "        if self.pc:\n",
    "            result += '<TR><TD ALIGN=\"CENTER\"><FONT FACE=\"Courier\">' + z3_html_escape(self.pc) + \"</FONT></TD></TR>\"\n",
    "        result += \"</TABLE>\"\n",
    "\n",
    "        z3.set_html_mode(prev_html)\n",
    "        return result\n",
    "\n",
    "    def __repr__(self):\n",
    "        return f\"SETNode({repr(self.environment)}, {repr(self.pc)})\"\n",
    "\n",
    "    def __eq__(self, other):\n",
    "        return type(other) is type(self) and self.environment == other.environment and self.pc == other.pc"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Symbolic execution trees themselves are defined just like derivation trees, only with `SETNode`s instead of strings as nodes."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "SET = Tuple[SETNode, List['SET']]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The following function produces graphical representations (using GraphViz) of SETs. Leaves are highlighted using the following color scheme: Red leaves represent raised exceptions, green leaves represent nodes with unsatisfiable path conditions, and blue leaves all other nodes containing \"normal\" computation results."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "from graphviz import Digraph"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "def display_set(symbolic_execution_tree: SET) -> Digraph:\n",
    "    def traverse_tree(dot: Digraph, tree: SET, node_id=0):\n",
    "        node, children = tree\n",
    "        \n",
    "        attrs = {}\n",
    "        if node.environment.unsatisfiable():\n",
    "            attrs = {\"fillcolor\": \"lightgreen\", \"style\": \"filled\"}\n",
    "        elif (node.environment.abrupt_completion is not None and\n",
    "              not isinstance(node.environment.abrupt_completion, AbruptCompletionNoException)):\n",
    "            attrs = {\"fillcolor\": \"lightpink1\", \"style\": \"filled\"}\n",
    "        elif not children:\n",
    "            attrs = {\"fillcolor\": \"paleturquoise2\", \"style\": \"filled\"}\n",
    "\n",
    "        dot.node(repr(node_id), \"<\" + node.to_html() + \">\", _attributes=attrs)\n",
    "\n",
    "        if children:\n",
    "            for child in children:\n",
    "                nonlocal counter\n",
    "                counter += 1\n",
    "                child_id = counter\n",
    "                dot.edge(repr(node_id), repr(child_id))\n",
    "                traverse_tree(dot, child, child_id)\n",
    "\n",
    "    counter = 0\n",
    "    dot = Digraph(comment=\"Symbolic Execution Tree\")\n",
    "    dot.attr('node', shape='box')\n",
    "    traverse_tree(dot, symbolic_execution_tree)\n",
    "    return dot"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Recall our small example from the previous section:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {
    "tags": [
     "remove-cell"
    ]
   },
   "outputs": [],
   "source": [
    "x = 0\n",
    "y = 0"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {
    "tags": [
     "minipy"
    ]
   },
   "outputs": [],
   "source": [
    "x = 2 * y\n",
    "if x < 0:\n",
    "    x = -x"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The final SET resulting from symbolically executing this program will contain two leaves, one branch each for the case where the condition of the `if` statements hold or does not hold, respectively. Apart from the root, it also contains two inner nodes corresponding to the first assignment and the `if` statement. In the absence of an implemented symbolic interpreter, we manually construct this SET:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "image/svg+xml": [
       "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n",
       "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"\n",
       " \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n",
       "<!-- Generated by graphviz version 2.43.0 (0)\n",
       " -->\n",
       "<!-- Title: %3 Pages: 1 -->\n",
       "<svg width=\"524pt\" height=\"308pt\"\n",
       " viewBox=\"0.00 0.00 523.50 308.00\" xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n",
       "<g id=\"graph0\" class=\"graph\" transform=\"scale(1 1) rotate(0) translate(4 304)\">\n",
       "<title>%3</title>\n",
       "<polygon fill=\"white\" stroke=\"transparent\" points=\"-4,4 -4,-304 519.5,-304 519.5,4 -4,4\"/>\n",
       "<!-- 0 -->\n",
       "<g id=\"node1\" class=\"node\">\n",
       "<title>0</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"358,-300 171,-300 171,-248 358,-248 358,-300\"/>\n",
       "<text text-anchor=\"start\" x=\"183.5\" y=\"-280.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {x: x, y: y})</text>\n",
       "<text text-anchor=\"start\" x=\"227.5\" y=\"-259.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">x = 2 * y</text>\n",
       "</g>\n",
       "<!-- 1 -->\n",
       "<g id=\"node2\" class=\"node\">\n",
       "<title>1</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"363.5,-212 165.5,-212 165.5,-160 363.5,-160 363.5,-212\"/>\n",
       "<text text-anchor=\"start\" x=\"177.5\" y=\"-192.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {x: 2·y, y: y})</text>\n",
       "<text text-anchor=\"start\" x=\"227\" y=\"-171.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">if x &lt; 0:</text>\n",
       "</g>\n",
       "<!-- 0&#45;&gt;1 -->\n",
       "<g id=\"edge1\" class=\"edge\">\n",
       "<title>0&#45;&gt;1</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M264.5,-247.66C264.5,-239.7 264.5,-230.76 264.5,-222.24\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"268,-222.16 264.5,-212.16 261,-222.16 268,-222.16\"/>\n",
       "</g>\n",
       "<!-- 2 -->\n",
       "<g id=\"node3\" class=\"node\">\n",
       "<title>2</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"261,-124 2,-124 2,-72 261,-72 261,-124\"/>\n",
       "<text text-anchor=\"start\" x=\"14.5\" y=\"-104.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(0 ≤ x)}, None, {x: 2·y, y: y})</text>\n",
       "<text text-anchor=\"start\" x=\"107\" y=\"-83.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">x = &#45;x</text>\n",
       "</g>\n",
       "<!-- 1&#45;&gt;2 -->\n",
       "<g id=\"edge2\" class=\"edge\">\n",
       "<title>1&#45;&gt;2</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M225.67,-159.89C211.19,-150.53 194.57,-139.78 179.37,-129.95\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"180.9,-126.78 170.61,-124.29 177.1,-132.66 180.9,-126.78\"/>\n",
       "</g>\n",
       "<!-- 4 -->\n",
       "<g id=\"node5\" class=\"node\">\n",
       "<title>4</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"515.5,-116 279.5,-116 279.5,-80 515.5,-80 515.5,-116\"/>\n",
       "<text text-anchor=\"start\" x=\"291.5\" y=\"-93.8\" font-family=\"Times,serif\" font-size=\"14.00\">({0 ≤ x}, None, {x: 2·y, y: y})</text>\n",
       "</g>\n",
       "<!-- 1&#45;&gt;4 -->\n",
       "<g id=\"edge4\" class=\"edge\">\n",
       "<title>1&#45;&gt;4</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M303.33,-159.89C322.04,-147.79 344.31,-133.39 362.5,-121.63\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"364.57,-124.46 371.06,-116.1 360.76,-118.59 364.57,-124.46\"/>\n",
       "</g>\n",
       "<!-- 3 -->\n",
       "<g id=\"node4\" class=\"node\">\n",
       "<title>3</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"263,-36 0,-36 0,0 263,0 263,-36\"/>\n",
       "<text text-anchor=\"start\" x=\"12.5\" y=\"-13.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(0 ≤ x)}, None, {x: &#45;2·y, y: y})</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;3 -->\n",
       "<g id=\"edge3\" class=\"edge\">\n",
       "<title>2&#45;&gt;3</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M131.5,-71.89C131.5,-63.84 131.5,-54.88 131.5,-46.66\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"135,-46.41 131.5,-36.41 128,-46.41 135,-46.41\"/>\n",
       "</g>\n",
       "</g>\n",
       "</svg>\n"
      ],
      "text/plain": [
       "<graphviz.dot.Digraph at 0x7fb03cac3af0>"
      ]
     },
     "execution_count": 9,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "x = Variable(\"x\", INT_TYPE)\n",
    "y = Variable(\"y\", INT_TYPE)\n",
    "\n",
    "env_0 = SymbolicEnvironment(SymbolicStore({x: x.to_z3(), y: y.to_z3()}))\n",
    "env_1 = env_0.set(x, z3.IntVal(2) * y.to_z3())\n",
    "env_2 = env_1.add_constraint(x.to_z3() < z3.IntVal(0))\n",
    "env_2_1 = env_2.set(x, z3.IntVal(-1) * x.to_z3())\n",
    "env_3 = env_1.add_constraint(z3.Not(x.to_z3() < z3.IntVal(0)))\n",
    "\n",
    "root = SETNode(env_0, \"x = 2 * y\")\n",
    "if_node = SETNode(env_1, \"if x < 0:\")\n",
    "inversion_node = SETNode(env_2, \"x = -x\")\n",
    "leave_1 = SETNode(env_2_1)\n",
    "leave_2 = SETNode(env_3)\n",
    "\n",
    "tree = (root, [\n",
    "    (if_node, [\n",
    "        (inversion_node, [(leave_1, [])]),\n",
    "        (leave_2, [])])])\n",
    "\n",
    "display_set(tree)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Evaluating Expressions\n",
    "\n",
    "Also the symbolic evaluation of expressions can have multiple results. Consider the expression `y // x`: Depending on whether `x` is zero or non-zero, evaluating this expression results in a value or a `ZeroDivisionError`. Thus, our result type for expression evaluation is a collection of pairs of either a value or an exception, together with a boolean constraint signaling under which condition the respective result is returned."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "EvalExprResult = List[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We define a helper function `handle_binary_operation` computing all conditioned results for a binary operation (e.g., division or conjunction) applied on two evaluation results."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [],
   "source": [
    "def handle_binary_operation(left: EvalExprResult, right: EvalExprResult, op: str) -> EvalExprResult:\n",
    "    result = []\n",
    "\n",
    "    for left_result in left:\n",
    "        left_val, left_constr = left_result\n",
    "\n",
    "        if isinstance(left_val, Exception):\n",
    "            result.append(left_result)\n",
    "            continue\n",
    "\n",
    "        if op == \"and\":\n",
    "            result.append((z3.BoolVal(False), z3.simplify(z3.And(left_constr, z3.Not(left_val)))))\n",
    "        if op == \"or\":\n",
    "            result.append((z3.BoolVal(True), z3.simplify(z3.And(left_constr, left_val))))\n",
    "\n",
    "        for right_result in right:\n",
    "            right_val, right_constr = right_result\n",
    "            constraint_conjunction = z3.And(left_constr, right_constr)\n",
    "            constraint_conjunction = z3_and(z3.simplify(constraint_conjunction))\n",
    "\n",
    "            if z3.is_false(constraint_conjunction):\n",
    "                continue\n",
    "\n",
    "            if isinstance(right_val, Exception):\n",
    "                # Short-circuit evaluation:\n",
    "                if op == \"or\":\n",
    "                    constraint_conjunction = z3.And(constraint_conjunction, z3.Not(left_val))\n",
    "                elif op == \"and\":\n",
    "                    constraint_conjunction = z3.And(constraint_conjunction, left_val)\n",
    "\n",
    "                constraint_conjunction = z3.simplify(constraint_conjunction)\n",
    "                if not z3.is_false(constraint_conjunction):\n",
    "                    result.append((right_val, constraint_conjunction))\n",
    "\n",
    "                continue\n",
    "\n",
    "            left_val: Union[z3.ArithRef, z3.BoolRef]\n",
    "            right_val: Union[z3.ArithRef, z3.BoolRef]\n",
    "\n",
    "            if op == \"and\":\n",
    "                constraint_conjunction = z3.simplify(z3.And(constraint_conjunction, left_val))\n",
    "                if not z3.is_false(constraint_conjunction):\n",
    "                    result.append((right_val, constraint_conjunction))\n",
    "                continue\n",
    "\n",
    "            if op == \"or\":\n",
    "                constraint_conjunction = z3.simplify(z3.And(constraint_conjunction, z3.Not(left_val)))\n",
    "                if not z3.is_false(constraint_conjunction):\n",
    "                    result.append((right_val, constraint_conjunction))\n",
    "                continue\n",
    "\n",
    "            operator_map = {\n",
    "                \"+\": lambda a, b: z3.Concat(a, b) if isinstance(a, z3.SeqRef) else a + b,\n",
    "                \"-\": operator.sub,\n",
    "                \"*\": operator.mul,\n",
    "                \"==\": operator.eq,\n",
    "                \"!=\": lambda a, b: z3.Not(a == b),\n",
    "                \"<\": operator.lt,\n",
    "                \"<=\": operator.le,\n",
    "                \">\": operator.gt,\n",
    "                \">=\": operator.ge,\n",
    "                # \"and\": lambda a, b: z3.And(a, b),\n",
    "                # \"or\": lambda a, b: z3.Or(a, b)\n",
    "            }\n",
    "\n",
    "            if op in operator_map:\n",
    "                result.append((operator_map[op](left_val, right_val), constraint_conjunction))\n",
    "                continue\n",
    "\n",
    "            right_zero = right_val == z3.IntVal(0)\n",
    "            right_nonzero = z3.Not(right_zero)\n",
    "\n",
    "            if op == \"//\":\n",
    "                result.append((z3.ToInt(z3.ToReal(left_val) / z3.ToReal(right_val)),\n",
    "                               z3.And(constraint_conjunction, right_nonzero)))\n",
    "                result.append((ZeroDivisionError(\"division by zero\"),\n",
    "                               z3.And(constraint_conjunction, right_zero)))\n",
    "                continue\n",
    "\n",
    "            if op == \"%\":\n",
    "                result.append((left_val % right_val, z3.And(constraint_conjunction, right_nonzero)))\n",
    "                result.append((ZeroDivisionError(\"division by zero\"),\n",
    "                               z3.And(constraint_conjunction, right_zero)))\n",
    "                continue\n",
    "\n",
    "            assert False\n",
    "\n",
    "    return result"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "[(ToInt(ToReal(x)/ToReal(y)), And(True, Not(0 == y))), (ZeroDivisionError('division by zero'), And(True, 0 == y))]"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "x = Variable(\"x\", INT_TYPE)\n",
    "y = Variable(\"y\", INT_TYPE)\n",
    "\n",
    "result = handle_binary_operation(\n",
    "    [(x.to_z3(), z3.BoolVal(True))],\n",
    "    [(y.to_z3(), z3.BoolVal(True))],\n",
    "    \"//\")\n",
    "\n",
    "display_html(str(result))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The term produced for the division, `ToInt(ToReal(x)/ToReal(y))`, might come unexpected: Why not simply `x / y`? The reason is the semantics of Python's Integer division `//`, which is a *floor division* [^floor-division]. Essentially, Python rounds towards 0 for positive, and away from 0 for negative numbers. That is, `1 // -2` evaluates to -1, while `z3.IntVal(1) / z3.IntVal(-2)` evaluates to *0*. Since the `z3.ToInt` function rounds away from zero for negative numbers, the implementation above correctly models Python's (and minipy's) floor division semantics."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Based on this function, the implementation of expression evaluation can be handled uniformly in many cases. For example, consider the evaluation of sums and terms below:"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def evaluate_sum(self, expr: Sum, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        left = self.evaluate(expr.left, environment)\n",
    "        right = self.evaluate(expr.right, environment)\n",
    "        return handle_binary_operation(left, right, expr.op)\n",
    "\n",
    "    def evaluate_term(self, expr: Term, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        left = self.evaluate(expr.left, environment)\n",
    "        right = self.evaluate(expr.right, environment)\n",
    "        return handle_binary_operation(left, right, expr.op)\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "When evaluating tuple expressions, we sequentially evaluate all tuple elements and add an exception result if one of these evaluations raises an exception. We create a z3 sequence expression for the case where all tuple elements evaluate to a value. Consider the tuple `(x // y, x // (y + 1))`. Its symbolic evaluation has three results:\n",
    "\n",
    "1. y is 0, then the tuple evaluates to a `ZeroDivisionError`.\n",
    "2. y is -1, then the tuple evaluates to a `ZeroDivisionError`.\n",
    "3. y does not equal either 0 or -1, then the tuple evaluates to a sequence expression containing the symbolic values corresponding to the division.\n",
    "\n",
    "For considering all these options, we build the product of the individual evaluation results for the tuple elements. This yields a list of tuples, where each tuple has the length of the evaluated tuple and corresponds to one evaluation result. For example:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [],
   "source": [
    "import itertools"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[((ToInt(ToReal(x)/ToReal(y)), And(True, Not(0 == y))),\n",
       "  (ToInt(ToReal(x)/ToReal(y + 1)), And(True, Not(0 == y + 1)))),\n",
       " ((ToInt(ToReal(x)/ToReal(y)), And(True, Not(0 == y))),\n",
       "  (ZeroDivisionError('division by zero'), And(True, 0 == y + 1))),\n",
       " ((ZeroDivisionError('division by zero'), And(True, 0 == y)),\n",
       "  (ToInt(ToReal(x)/ToReal(y + 1)), And(True, Not(0 == y + 1)))),\n",
       " ((ZeroDivisionError('division by zero'), And(True, 0 == y)),\n",
       "  (ZeroDivisionError('division by zero'), And(True, 0 == y + 1)))]"
      ]
     },
     "execution_count": 14,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "result_1 = handle_binary_operation(\n",
    "    [(x.to_z3(), z3.BoolVal(True))],\n",
    "    [(y.to_z3(), z3.BoolVal(True))],\n",
    "    \"//\")\n",
    "\n",
    "result_2 = handle_binary_operation(\n",
    "    [(x.to_z3(), z3.BoolVal(True))],\n",
    "    [(y.to_z3() + z3.IntVal(1), z3.BoolVal(True))],\n",
    "    \"//\")\n",
    "\n",
    "results: List[EvalExprResult] = [result_1, result_2]\n",
    "list(itertools.product(*results))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We loop over the elements of the tuples in the list, and stop at the first exception (if any), which will be the evaluation result. Along the way, we collect the path constraints of the seen evaluation results for the tuple elements. Note that the list above contains four, and not three elements. When considering those, the third and fourth evaluation alternatives collapse to a single one, since we stop after processing the evaluation result for the first tuple element, which is an exception (the fourth element is anyway infeasible, since `y` cannot attain the values 0 and -1 at the same time)."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "To avoid creating the same result for equivalent evaluations (e.g., if the same exception is raised under the same condition when evaluating different tuple elements), we first collect evaluation results in a set. We wrap exceptions inside a wrapper object computing sensible hashes as Python exceptions generally do not supply a suitable hashing implementation."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def evaluate_tuple(self, expr: TupleNode, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        sub_results: List[EvalExprResult] = [self.evaluate(elem, environment) \n",
    "                                             for elem in expr.elems]\n",
    "        result: OrderedSet[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]] = OrderedSet([])\n",
    "\n",
    "        combinations: List[Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]] = \\\n",
    "            list(itertools.product(*sub_results))\n",
    "\n",
    "        combination: Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]\n",
    "        for combination in combinations:\n",
    "            values: List[SymbolicValueType] = []\n",
    "            path_constraint: z3.BoolRef = z3.BoolVal(True)\n",
    "            sub_result: Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]\n",
    "            for sub_value, sub_constraint in combination:\n",
    "                path_constraint = z3.simplify(z3.And(path_constraint, sub_constraint))\n",
    "                if isinstance(sub_value, Exception):\n",
    "                    result.add(wrap_exceptions((sub_value, path_constraint)))\n",
    "                    break\n",
    "\n",
    "                values.append(sub_value)\n",
    "\n",
    "            if len(values) != len(combination):\n",
    "                continue\n",
    "\n",
    "            if len(values) == 1:\n",
    "                value = values[0]\n",
    "                result.add((value if isinstance(value.sort(), z3.SeqSortRef) \n",
    "                            else z3.Unit(value), path_constraint))\n",
    "            elif len(values) > 1:\n",
    "                result.add((z3.Concat(*[value if isinstance(value.sort(), z3.SeqSortRef) \n",
    "                                        else z3.Unit(value)\n",
    "                                        for value in values]), path_constraint))\n",
    "\n",
    "        return [unwrap_exceptions(r) for r in result]\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "For evaluating tuple access (e.g., `t[x]`), we have to consider the possible exceptions that may arise: Apart form the exceptions raised by evaluating `t` and `x` (which can be non-atomic expressions), `x` might be out of range, or `t` and `x` might have the wrong sorts (i.e., not `tuple` or `int`). All this is considered in the following implementation:"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def handle_tuple_access(self, tuple_result: EvalExprResult, expr_result: EvalExprResult) -> EvalExprResult:\n",
    "        result: EvalExprResult = []\n",
    "\n",
    "        for atom_value, atom_constraint in tuple_result:\n",
    "            if isinstance(atom_value, Exception):\n",
    "                result.append((atom_value, atom_constraint))\n",
    "                continue\n",
    "\n",
    "            if not isinstance(atom_value.sort(), z3.SeqSortRef):\n",
    "                result.append((TypeError(f\"'{atom_value.sort()} is not subscriptable\"),\n",
    "                               atom_constraint))\n",
    "                continue\n",
    "\n",
    "            for expr_value, expr_constraint in expr_result:\n",
    "                conjunction = z3.And(atom_constraint, expr_constraint)\n",
    "                if isinstance(expr_value, Exception):\n",
    "                    result.append((expr_value, conjunction))\n",
    "                    continue\n",
    "\n",
    "                if expr_value.sort() != z3.IntSort():\n",
    "                    result.append((TypeError(f\"list indices must be integers, not {expr_value.sort()}\"),\n",
    "                                   conjunction))\n",
    "                    continue\n",
    "\n",
    "                atom_value: z3.SeqRef\n",
    "\n",
    "                # In Python, indices can be in the closed interval [-len ; len - 1]\n",
    "                length_constraint_1 = z3.And(z3.IntVal(0) <= expr_value,\n",
    "                                             expr_value < z3.Length(atom_value))  # [0 ; len - 1]\n",
    "                length_constraint_2 = z3.And(z3.IntVal(0) > expr_value,\n",
    "                                             expr_value >= z3.IntVal(-1) * z3.Length(atom_value))  # [-len ; -1]\n",
    "\n",
    "                expr = z3.simplify(atom_value[expr_value])\n",
    "                if expr.decl().kind() == z3.Z3_OP_ITE and \\\n",
    "                        implies(z3.And(conjunction, length_constraint_1), expr.children()[0]):\n",
    "                    expr = expr.children()[1]\n",
    "\n",
    "                result.append((expr, z3.And(conjunction, length_constraint_1)))\n",
    "\n",
    "                expr = z3.simplify(atom_value[z3.Length(atom_value) + expr_value])\n",
    "                if expr.decl().kind() == z3.Z3_OP_ITE and \\\n",
    "                        implies(z3.And(conjunction, length_constraint_2), expr.children()[0]):\n",
    "                    expr = expr.children()[1]\n",
    "\n",
    "                result.append((expr, z3.And(conjunction, length_constraint_2)))\n",
    "\n",
    "                result.append((IndexError(\"tuple index out of range\"),\n",
    "                               z3.And(conjunction, z3.Not(length_constraint_1), z3.Not(length_constraint_2))))\n",
    "\n",
    "        return result\n",
    "\n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The execution of function definitions adds a `Callable` to the symbolic environment, which accepts an environment, a list of symbolic arguments, and returns an `EvalExprResult`. The evaluation of a function call has to consider all exceptions thrown when evaluating the arguments. Additionally, an exception is registered if an unexpected number of arguments is given. Note that we do not check argument types, which is generally not done in a dynamically typed language like Python. In minipy, however, we always have complete type information, which is why this problem could be ruled out by a static check taking place before symbolic execution."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def evaluate_fun_call(self, expr: FunctionCall, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        result: EvalExprResult = []\n",
    "\n",
    "        f_name = expr.name\n",
    "        if f_name not in environment.functions:\n",
    "            raise NameError(f\"name '{f_name}' is not defined\")\n",
    "\n",
    "        args_result = [self.evaluate(arg, environment) for arg in expr.args]\n",
    "\n",
    "        combinations: List[Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]] = \\\n",
    "            list(itertools.product(*args_result))\n",
    "\n",
    "        combination: Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]\n",
    "        for combination in combinations:\n",
    "            values: List[SymbolicValueType] = []\n",
    "            path_constraint: z3.BoolRef = z3.BoolVal(True)\n",
    "            sub_result: Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]\n",
    "            for sub_value, sub_constraint in combination:\n",
    "                path_constraint = z3.simplify(z3.And(path_constraint, sub_constraint))\n",
    "                if isinstance(sub_value, Exception):\n",
    "                    result.append(wrap_exceptions((sub_value, path_constraint)))\n",
    "                    break\n",
    "\n",
    "                values.append(sub_value)\n",
    "\n",
    "            if len(values) != len(combination):\n",
    "                continue\n",
    "\n",
    "            if f_name in environment.functions:\n",
    "                params, ret_type, fun = environment.functions[f_name]\n",
    "                if len(values) != len(params):\n",
    "                    raise TypeError(\n",
    "                        f\"{f_name}({', '.join(map(str, params))}) takes {len(params)} \"\n",
    "                        f\"argument(s) but {len(values)} were given\")\n",
    "\n",
    "                result.extend(fun(environment, *values))\n",
    "            else:\n",
    "                assert False\n",
    "\n",
    "        return result\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Executing Statements"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Similarly to the evaluation of expressions, the execution of statements has to consider all possible outcomes, such that the execution result is a tree and not a mere update of the environment. We explain the implementation of the execution functions for the stereotypical assignment, `if`, `try`, and `while` statements; moreover, we have a look at the function definitions and calls.\n",
    "\n",
    "Executing an assignment results in one leaf node per raised exception during the evaluation of the right-hand side, plus a normal evaluation result when the right-hand side evaluates to a value. "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def execute_assignment(self, stmt: Assignment, environment: SymbolicEnvironment) -> SET:\n",
    "        eval_expr_result = [\n",
    "            (value, constraint)\n",
    "            for value, constraint in self.evaluate(stmt.expression, environment)\n",
    "            if not self.unsatisfiable(environment.add_constraint(constraint))\n",
    "        ]\n",
    "\n",
    "        children: List[SET] = []\n",
    "        result = (SETNode(environment, stmt.code), children)\n",
    "\n",
    "        for value, constraint in eval_expr_result:\n",
    "            if isinstance(value, Exception):\n",
    "                new_env = environment.set_abrupt_completion(value)\n",
    "            else:\n",
    "                variable = Variable(stmt.lhs, get_type_for_z3_expr(value))\n",
    "                new_env = environment.set(variable, value)\n",
    "\n",
    "            new_env = new_env.add_constraint(constraint)\n",
    "\n",
    "            children.append((SETNode(new_env, None), []))\n",
    "        \n",
    "        return result\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "In the case of an `if` statement, we have one leaf for each raised exception when evaluating the guard, plus one leaf each for the execution of the then and else branch of the conditional. To avoid creating infeasible subtrees, we add an unsatisfiability check for each case. To that end, we add a method `constraint_unsatisfiable` (and a convenience method `unsatisfiable` for environments) to `SymbolicInterpreter`, which, as of now, only works as a proxy to `is_unsat`. Yet, it will greatly simplify adapting the symbolic to a concrete interpreter in {ref}`techniques:concolic_execution`."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def constraint_unsatisfiable(self, constraint: z3.BoolRef) -> bool:\n",
    "        return is_unsat(constraint)\n",
    "\n",
    "    def unsatisfiable(self, env: SymbolicEnvironment) -> bool:\n",
    "        return self.constraint_unsatisfiable(z3_and(*env.path_constraints))\n",
    "    \n",
    "    def execute_if_stmt(self, stmt: IfStmt, environment: SymbolicEnvironment) -> SET:\n",
    "        eval_expr_result = self.evaluate(stmt.guard, environment)\n",
    "\n",
    "        result: SET = (SETNode(environment, stmt.code), [])\n",
    "\n",
    "        for value, constraint in eval_expr_result:\n",
    "            new_env = environment.add_constraint(constraint)\n",
    "\n",
    "            if self.unsatisfiable(new_env):\n",
    "                continue\n",
    "\n",
    "            if isinstance(value, Exception):\n",
    "                result[1].append((SETNode(new_env.set_abrupt_completion(value), ''), []))\n",
    "                continue\n",
    "\n",
    "            assert z3.is_bool(value)\n",
    "\n",
    "            then_constraint: z3.BoolRef = z3.simplify(value)\n",
    "            else_constraint = z3.simplify(z3.Not(value))\n",
    "\n",
    "            then_env = new_env.add_constraint(then_constraint)\n",
    "            else_env = new_env.add_constraint(else_constraint)\n",
    "\n",
    "            then_env_unsat = self.unsatisfiable(then_env)\n",
    "            else_env_unsat = self.unsatisfiable(else_env)\n",
    "            \n",
    "            assert not then_env_unsat or not else_env_unsat\n",
    "\n",
    "            if not then_env_unsat:\n",
    "                subtree = self.execute(stmt.then_block, then_env)\n",
    "                result[1].append(subtree)\n",
    "\n",
    "            if not else_env_unsat:\n",
    "                if stmt.else_block is not None:\n",
    "                    subtree = self.execute(stmt.else_block, else_env)\n",
    "                    result[1].append(subtree)\n",
    "                else:\n",
    "                    result[1].append((SETNode(else_env, ''), []))\n",
    "\n",
    "        return result\n",
    "\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The path condition for the then leg of the `if` statement is extended by the constraint corresponding to the evaluation of the guard along with the evaluation result of the guard expression. Including the constraint for the value of the guard is important if its evaluation can raise an exception; then, we may only execute the bodies of the `if` statement in the normal completion case.\n",
    "\n",
    "Executing a `try` statement is relatively simple. The only specialty is that we have to look up exception types from `except` clauses to check whether a raised exception matches the caught type (alternatively, we could resort to simple class name comparisons)."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def execute_try_stmt(self, stmt: TryStmt, environment: Environment) -> None:\n",
    "        try:\n",
    "            self.execute(stmt.block, environment)\n",
    "        except Exception as exc:\n",
    "            if issubclass(type(exc), AbruptCompletionNoException):\n",
    "                raise exc\n",
    "\n",
    "            if stmt.exc_type is not None:\n",
    "                caught_exc_type = getattr(sys.modules[\"builtins\"], stmt.exc_type)\n",
    "                if issubclass(type(exc), caught_exc_type):\n",
    "                    self.execute(stmt.except_block, environment)\n",
    "                    return\n",
    "            else:\n",
    "                self.execute(stmt.except_block, environment)\n",
    "                return\n",
    "\n",
    "            assert False\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The most \"problematic\" statement in symbolic execution of minipy is the `while` statement. Consider the implementation of the `execute_while_stmt` method in the concrete minipy interpreter:"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class Interpreter:\n",
    "    # ...\n",
    "    \n",
    "    def execute_while_stmt(self, stmt: WhileStmt, environment: Environment) -> None:\n",
    "        eval_guard_res = self.evaluate(stmt.guard, environment)\n",
    "        assert type(eval_guard_res) is bool\n",
    "\n",
    "        while eval_guard_res:\n",
    "            try:\n",
    "                self.execute(stmt.body, environment)\n",
    "            except Continue:\n",
    "                eval_guard_res = self.evaluate(stmt.guard, environment)\n",
    "                continue\n",
    "            except Break:\n",
    "                break\n",
    "            eval_guard_res = self.evaluate(stmt.guard, environment)\n",
    "        else:\n",
    "            if stmt.else_block is not None:\n",
    "                self.execute(stmt.else_block, environment)\n",
    "                return\n",
    "\n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The concrete interpreter evaluates the guard, executes the body if the guard evaluates to true, checks for the special `Continue` and `Break` exceptions, and continues doing this until the guard evaluates to false (or a `Continue` or `Break` was encountered). The simplest way of symbolically executing a loop is to mimic this behavior, which is called *loop unrolling* in SE terminology {cite}`ahrendt.beckert.ea-16` {cite}`baldoni.coppa.ea-18`. Implementing this is not too difficult:"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def execute_while_stmt(self, stmt: WhileStmt, environment: SymbolicEnvironment) -> SET:\n",
    "        eval_guard_res = self.evaluate(stmt.guard, environment)\n",
    "        result = (SETNode(environment, stmt.code), [])\n",
    "\n",
    "        for value, constraint in eval_guard_res:\n",
    "            new_env = environment.add_constraint(constraint)\n",
    "            if self.unsatisfiable(new_env):\n",
    "                continue\n",
    "            \n",
    "            if isinstance(value, Exception):\n",
    "                exc_node = SETNode(new_env.set_abrupt_completion(value), '')\n",
    "                result[1].append((exc_node, []))\n",
    "                continue\n",
    "\n",
    "            then_constraint: z3.BoolRef = z3.simplify(value)\n",
    "            then_env = new_env.add_constraint(then_constraint)\n",
    "\n",
    "            if not self.unsatisfiable(then_env):\n",
    "                subtree = self.execute(stmt.body, then_env)\n",
    "                result[1].append(subtree)\n",
    "                for leaf in get_leaves(subtree):\n",
    "                    node, _ = leaf\n",
    "                    env = node.environment\n",
    "                    exc = env.abrupt_completion\n",
    "                    if exc is not None:\n",
    "                        if isinstance(exc, Continue):\n",
    "                            env = env.set_abrupt_completion(None)\n",
    "                        elif isinstance(exc, Break):\n",
    "                            env = env.set_abrupt_completion(None)\n",
    "                            result = replace_in_tree(\n",
    "                                result, leaf, (SETNode(env, ''), []))\n",
    "                            continue\n",
    "                        else:\n",
    "                            continue\n",
    "\n",
    "                    while_tree = self.execute(stmt, env)\n",
    "                    result = replace_in_tree(result, leaf, while_tree)\n",
    "                    continue\n",
    "\n",
    "            else_constraint: z3.BoolRef = z3.simplify(z3.Not(value))\n",
    "            else_env = new_env.add_constraint(else_constraint)\n",
    "\n",
    "            if not self.unsatisfiable(else_env):\n",
    "                if stmt.else_block is not None:\n",
    "                    subtree = self.execute(stmt.else_block, else_env)\n",
    "                    result[1].append((SETNode(else_env, 'else:'), [subtree]))\n",
    "                else:\n",
    "                    result[1].append((SETNode(else_env, ''), []))\n",
    "\n",
    "        return result\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Apart from the standard differences between the concrete and symbolic interpreter (which correspond to the differences of the respective implementations of the execution of an `if` statement), the most striking difference is that the implementation is recursive, which is a little easier to implement. Furthermore, the execution of the body can, of course, result in a tree with multiple leaves. We replace each leaf from that tree with a leaf indicating what to execute next. For instance, in the case of an encountered `Break`, we replace the leaf node with a single node without children, and reset the abrupt completion flag (`env = env.set_abrupt_completion(None)`) since the loop \"catches\" abrupt completion due to `break` statements. For the normal completion case, we execute the loop once again, and replace the normal completion leaves of the body with the newly generated trees resulting from the loop execution."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The execution of function definition statements in the symbolic interpreter works analogously to the concrete minipy interpreter: We add a continuation to the execution environment which is used when evaluating later calls to the defined functions. The corresponding implementation in the concrete interpreter is "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class Interpreter:\n",
    "    \n",
    "    # ...\n",
    "    \n",
    "    def execute_function_def(self, stmt: FunctionDef, environment: Environment) -> None:\n",
    "        f_name = stmt.name\n",
    "        ret_type = get_type(stmt.t)\n",
    "        params = [self.evaluate_param(param) for param in stmt.params]\n",
    "\n",
    "        def func(*args):\n",
    "            new_env = environment.copy()\n",
    "            for idx, parameter in enumerate(params):\n",
    "                new_env[parameter] = args[idx]\n",
    "            try:\n",
    "                self.execute_block(stmt.block, new_env)\n",
    "            except Return as ret:\n",
    "                return ret.value\n",
    "\n",
    "        environment[f_name] = (params, ret_type, func)\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The main difference in the symbolic variant is that it has to account for multiple leaves in the symbolic tree resulting from the execution of the function body:"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def execute_function_def(self, stmt: FunctionDef, environment: SymbolicEnvironment) -> SET:\n",
    "        f_name = stmt.name\n",
    "        ret_type = get_type(stmt.t)\n",
    "        params = tuple([self.evaluate_param(param) for param in stmt.params])\n",
    "\n",
    "        def func(environment: SymbolicEnvironment, *args: SymbolicValueType) -> EvalExprResult:\n",
    "            for idx, parameter in enumerate(params):\n",
    "                environment = environment.set(parameter, args[idx])\n",
    "\n",
    "            block_result = self.execute(stmt.block, environment)\n",
    "            result: EvalExprResult = []\n",
    "\n",
    "            for node, _ in get_leaves(block_result):\n",
    "                exc = node.environment.abrupt_completion\n",
    "                if exc is None:\n",
    "                    raise RuntimeError(\"Missing return statement\")\n",
    "\n",
    "                if isinstance(exc, SymbolicReturn):\n",
    "                    result.append((exc.value, z3_and(*node.environment.path_constraints)))\n",
    "                else:\n",
    "                    result.append((exc, z3_and(*node.environment.path_constraints)))\n",
    "\n",
    "            return result\n",
    "\n",
    "        return SETNode(environment, stmt.code), [\n",
    "            (SETNode(environment.add_function(f_name, params, ret_type, func), ''), [])\n",
    "        ]\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Function *calls* are evaluated by first considering all evaluation alternatives for the passed parameters: If the evaluation of a parameter raises an exception, the exception is the result of the call. Only if all parameters evaluate normally, the function body is executed. Similarly to the evaluation of tuple expressions, we build the product of the evaluation results for the parameters to account for all evaluation alternatives."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "class SymbolicInterpreter:\n",
    "    # ...\n",
    "    \n",
    "    def evaluate_fun_call(self, expr: FunctionCall, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        result: EvalExprResult = []\n",
    "\n",
    "        f_name = expr.name\n",
    "        if f_name not in environment.functions:\n",
    "            raise NameError(f\"name '{f_name}' is not defined\")\n",
    "\n",
    "        args_result = [self.evaluate(arg, environment) for arg in expr.args]\n",
    "\n",
    "        combinations: List[Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]] = \\\n",
    "            list(itertools.product(*args_result))\n",
    "\n",
    "        combination: Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]\n",
    "        for combination in combinations:\n",
    "            values: List[SymbolicValueType] = []\n",
    "            path_constraint: z3.BoolRef = z3.BoolVal(True)\n",
    "            sub_result: Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]\n",
    "            for sub_value, sub_constraint in combination:\n",
    "                path_constraint = z3.simplify(z3.And(path_constraint, sub_constraint))\n",
    "                if isinstance(sub_value, Exception):\n",
    "                    result.append(wrap_exceptions((sub_value, path_constraint)))\n",
    "                    break\n",
    "\n",
    "                values.append(sub_value)\n",
    "\n",
    "            if len(values) != len(combination):\n",
    "                continue\n",
    "\n",
    "            if f_name in environment.functions:\n",
    "                params, ret_type, fun = environment.functions[f_name]\n",
    "                if len(values) != len(params):\n",
    "                    raise TypeError(f\"{f_name}({', '.join(map(str, params))}) takes {len(params)} \"\n",
    "                                    f\"argument(s) but {len(values)} were given\")\n",
    "\n",
    "                result.extend(fun(environment, *values))\n",
    "            else:\n",
    "                assert False\n",
    "\n",
    "        return result\n",
    "    \n",
    "    # ...\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The complete code for the symbolic interpreter, along with some helper functions, follows bellow (press the toggle on the right)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import logging\n",
    "from typing import Any\n",
    "from orderedset import OrderedSet\n",
    "\n",
    "\n",
    "SomeTree = Tuple[Any, List['SomeTree']]\n",
    "\n",
    "\n",
    "def tree_contains(tree: SET, needle: SET) -> bool:\n",
    "    if tree is needle:\n",
    "        return True\n",
    "    node, children = tree\n",
    "    if not children:\n",
    "        return False\n",
    "    return any(tree_contains(child, needle) for child in children)\n",
    "\n",
    "\n",
    "def replace_in_tree(tree: SET, to_replace: SET, new_tree: SET) -> SET:\n",
    "    assert tree_contains(tree, to_replace)\n",
    "    for path, subtree in path_iterator(tree):\n",
    "        if subtree is to_replace:\n",
    "            return replace_tree_path(tree, path, new_tree)\n",
    "        \n",
    "        \n",
    "def replace_tree_path(in_tree: SomeTree, path: Path, replacement_tree: SomeTree) -> SomeTree:\n",
    "    \"\"\"In-place subtree replacement preserving identity of other subtrees.\"\"\"\n",
    "    if not path:\n",
    "        return replacement_tree\n",
    "\n",
    "    curr_parent = in_tree\n",
    "    last_idx = path[0]\n",
    "    curr_tree = curr_parent[1][last_idx]\n",
    "    path = path[1:]\n",
    "\n",
    "    while path:\n",
    "        last_idx = path[0]\n",
    "        path = path[1:]\n",
    "        curr_parent = curr_tree\n",
    "        curr_tree = curr_tree[1][last_idx]\n",
    "\n",
    "    curr_parent[1][last_idx] = replacement_tree\n",
    "\n",
    "    return in_tree\n",
    "\n",
    "        \n",
    "def path_iterator(tree: SomeTree, path: Path = ()) -> Generator[Tuple[Path, SomeTree], None, None]:\n",
    "    yield path, tree\n",
    "    if tree[1] is not None:\n",
    "        for i, child in enumerate(tree[1]):\n",
    "            yield from path_iterator(child, path + (i,))\n",
    "\n",
    "        \n",
    "def get_type_for_z3_expr(expr: z3.ExprRef) -> Type:\n",
    "    if z3.is_int(expr):\n",
    "        return INT_TYPE\n",
    "    elif z3.is_seq(expr):\n",
    "        return TUPLE_TYPE\n",
    "    elif z3.is_bool(expr):\n",
    "        return BOOL_TYPE\n",
    "\n",
    "    assert False\n",
    "    \n",
    "    \n",
    "def implies(premise: z3.BoolRef, conclusion: z3.BoolRef) -> bool:\n",
    "    return is_unsat(z3.And(premise, z3.Not(conclusion)))\n",
    "        \n",
    "        \n",
    "class SymbolicReturn(AbruptCompletionNoException):\n",
    "    def __init__(self, value: SymbolicValueType):\n",
    "        self.value = value\n",
    "\n",
    "\n",
    "class SymbolicInterpreter:\n",
    "    def __init__(self):\n",
    "        self.logger = logging.getLogger(type(self).__name__)\n",
    "        \n",
    "        self.stmt_interpretations: Dict[typing.Type[ASTNode], Callable[[ASTNode, SymbolicEnvironment], SET]] = {\n",
    "            Stmts: self.execute_stmts,\n",
    "            Pass: self.execute_pass,\n",
    "            BreakStmt: self.execute_break,\n",
    "            ContinueStmt: self.execute_continue,\n",
    "            ReturnStmt: self.execute_return_stmt,\n",
    "            Assert: self.execute_assert_stmt,\n",
    "            Assignment: self.execute_assignment,\n",
    "            Block: self.execute_block,\n",
    "            FunctionDef: self.execute_function_def,\n",
    "            TryStmt: self.execute_try_stmt,\n",
    "            WhileStmt: self.execute_while_stmt,\n",
    "            IfStmt: self.execute_if_stmt,\n",
    "        }\n",
    "\n",
    "        self.expr_interpretations: Dict[typing.Type[Expression], Callable[[ASTNode, SymbolicEnvironment],\n",
    "                                                                          EvalExprResult]] = {\n",
    "            Disjunction: self.evaluate_disjunction,\n",
    "            Conjunction: self.evaluate_conjunction,\n",
    "            Inversion: self.evaluate_inversion,\n",
    "            Comparison: self.evaluate_comparison,\n",
    "            Sum: self.evaluate_sum,\n",
    "            Term: self.evaluate_term,\n",
    "            Factor: self.evaluate_factor,\n",
    "            TupleNode: self.evaluate_tuple,\n",
    "            TupleAccess: self.evaluate_tuple_access,\n",
    "            FunctionCall: self.evaluate_fun_call,\n",
    "            Param: self.evaluate_param,\n",
    "            BooleanAtom: self.evaluate_boolean,\n",
    "            IntAtom: self.evaluate_int,\n",
    "            NameAtom: self.evaluate_name,\n",
    "        }\n",
    "    \n",
    "    def constraint_unsatisfiable(self, constraint: z3.BoolRef) -> bool:\n",
    "        return is_unsat(constraint)\n",
    "\n",
    "    def unsatisfiable(self, env: SymbolicEnvironment) -> bool:\n",
    "        return self.constraint_unsatisfiable(z3_and(*env.path_constraints))\n",
    "\n",
    "    def execute(self, stmt: ASTNode, environment: SymbolicEnvironment) -> SET:\n",
    "        self.logger.debug(\"Executing statement %s\", stmt.code)\n",
    "\n",
    "        t = type(stmt)\n",
    "        if isinstance(stmt, WhileStmt):\n",
    "            return self.execute_while_stmt(stmt, environment)\n",
    "\n",
    "        if t in self.stmt_interpretations:\n",
    "            return self.stmt_interpretations[t](stmt, environment)\n",
    "\n",
    "        if isinstance(stmt, Expression):\n",
    "            # Expression Statements\n",
    "            return SETNode(environment, stmt.code), [(SETNode(environment.add_constraint(constraint), None), [])\n",
    "                                                     for _, constraint in self.evaluate(stmt, environment)]\n",
    "\n",
    "        assert False\n",
    "\n",
    "    def execute_stmts(self, stmts: Stmts, environment: SymbolicEnvironment) -> SET:\n",
    "        return self.execute_sequence(tuple(stmts.stmts), environment)\n",
    "\n",
    "    def execute_sequence(self, stmts: Tuple[ASTNode, ...], environment: SymbolicEnvironment) -> SET:\n",
    "        result = None\n",
    "        leaves: List[SET] = []\n",
    "\n",
    "        for stmt in stmts:\n",
    "            if result is None:\n",
    "                tree = self.execute(stmt, environment)\n",
    "                result, leaves = tree, get_leaves(tree)\n",
    "                continue\n",
    "\n",
    "            for leaf in leaves:\n",
    "                node, children = leaf\n",
    "                if node.environment.abrupt_completion is not None or self.unsatisfiable(node.environment):\n",
    "                    continue\n",
    "                tree = self.execute(stmt, node.environment)\n",
    "\n",
    "                assert tree_contains(result, leaf)\n",
    "                result = replace_in_tree(result, leaf, tree)\n",
    "\n",
    "            leaves = get_leaves(result)\n",
    "\n",
    "        return result\n",
    "\n",
    "    def execute_pass(self, stmt: Pass, environment: SymbolicEnvironment) -> SET:\n",
    "        return SETNode(environment, stmt.code), [(SETNode(environment, None), [])]\n",
    "\n",
    "    def execute_break(self, stmt: BreakStmt, environment: SymbolicEnvironment) -> SET:\n",
    "        return SETNode(environment, stmt.code), [(SETNode(environment.set_abrupt_completion(Break()), None), [])]\n",
    "\n",
    "    def execute_continue(self, stmt: ContinueStmt, environment: SymbolicEnvironment) -> SET:\n",
    "        return SETNode(environment, stmt.code), [(SETNode(environment.set_abrupt_completion(Continue()), None), [])]\n",
    "\n",
    "    def execute_return_stmt(self, stmt: ReturnStmt, environment: SymbolicEnvironment) -> SET:\n",
    "        eval_expr_result = self.evaluate(stmt.expression, environment)\n",
    "\n",
    "        children: List[SET] = []\n",
    "        result = (SETNode(environment, stmt.code), children)\n",
    "\n",
    "        for value, constraint in eval_expr_result:\n",
    "            if isinstance(value, Exception):\n",
    "                children.append((SETNode(environment\n",
    "                                         .add_constraint(constraint)\n",
    "                                         .set_abrupt_completion(value), ''), []))\n",
    "            else:\n",
    "                children.append((SETNode(environment\n",
    "                                         .add_constraint(constraint)\n",
    "                                         .set_abrupt_completion(SymbolicReturn(value)), ''), []))\n",
    "\n",
    "        return result\n",
    "\n",
    "    def execute_assert_stmt(self, stmt: Assert, environment: SymbolicEnvironment) -> SET:\n",
    "        eval_expr_result = self.evaluate(stmt.expression, environment)\n",
    "\n",
    "        children: List[SET] = []\n",
    "        result = (SETNode(environment, stmt.code), children)\n",
    "\n",
    "        for value, constraint in eval_expr_result:\n",
    "            new_env = environment.add_constraint(constraint)\n",
    "            if self.unsatisfiable(new_env):\n",
    "                continue\n",
    "\n",
    "            if isinstance(value, Exception):\n",
    "                children.append((SETNode(new_env.set_abrupt_completion(value), ''), []))\n",
    "                continue\n",
    "                \n",
    "            assert isinstance(value, z3.BoolRef)\n",
    "            \n",
    "            sat_env = new_env.add_constraint(value)\n",
    "            viol_env = new_env.add_constraint(z3.Not(value))\n",
    "\n",
    "            sat_env_unsat = self.unsatisfiable(sat_env)\n",
    "            viol_env_unsat = self.unsatisfiable(viol_env)\n",
    "            \n",
    "            assert not sat_env_unsat or not viol_env_unsat\n",
    "\n",
    "            if sat_env_unsat:\n",
    "                children.append((SETNode(environment.set_abrupt_completion(AssertionError()), ''), []))\n",
    "            elif viol_env_unsat:\n",
    "                children.append((SETNode(environment, ''), []))\n",
    "            else:\n",
    "                children.append((SETNode(sat_env, ''), []))\n",
    "                children.append((SETNode(viol_env.set_abrupt_completion(AssertionError()), ''), []))\n",
    "\n",
    "        return result\n",
    "\n",
    "    def execute_assignment(self, stmt: Assignment, environment: SymbolicEnvironment) -> SET:\n",
    "        eval_expr_result = [\n",
    "            (value, constraint)\n",
    "            for value, constraint in self.evaluate(stmt.expression, environment)\n",
    "            if not self.unsatisfiable(environment.add_constraint(constraint))\n",
    "        ]\n",
    "\n",
    "        children: List[SET] = []\n",
    "        result = (SETNode(environment, stmt.code), children)\n",
    "\n",
    "        for value, constraint in eval_expr_result:\n",
    "            if isinstance(value, Exception):\n",
    "                new_env = environment.set_abrupt_completion(value)\n",
    "            else:\n",
    "                variable = Variable(stmt.lhs, get_type_for_z3_expr(value))\n",
    "                new_env = environment.set(variable, value)\n",
    "\n",
    "            new_env = new_env.add_constraint(constraint)\n",
    "\n",
    "            children.append((SETNode(new_env, None), []))\n",
    "        \n",
    "        return result\n",
    "\n",
    "    def execute_function_def(self, stmt: FunctionDef, environment: SymbolicEnvironment) -> SET:\n",
    "        f_name = stmt.name\n",
    "        ret_type = get_type(stmt.t)\n",
    "        params = tuple([self.evaluate_param(param) for param in stmt.params])\n",
    "\n",
    "        def func(environment: SymbolicEnvironment, *args: SymbolicValueType) -> EvalExprResult:\n",
    "            for idx, parameter in enumerate(params):\n",
    "                environment = environment.set(parameter, args[idx])\n",
    "\n",
    "            block_result = self.execute(stmt.block, environment)\n",
    "            result: EvalExprResult = []\n",
    "\n",
    "            for node, _ in get_leaves(block_result):\n",
    "                exc = node.environment.abrupt_completion\n",
    "                if exc is None:\n",
    "                    raise RuntimeError(\"Missing return statement\")\n",
    "\n",
    "                if isinstance(exc, SymbolicReturn):\n",
    "                    result.append((exc.value, z3_and(*node.environment.path_constraints)))\n",
    "                else:\n",
    "                    result.append((exc, z3_and(*node.environment.path_constraints)))\n",
    "\n",
    "            return result\n",
    "\n",
    "        return SETNode(environment, stmt.code), [\n",
    "            (SETNode(environment.add_function(f_name, params, ret_type, func), ''), [])\n",
    "        ]\n",
    "\n",
    "    def evaluate_param(self, param: Param) -> Variable:\n",
    "        return Variable(param.name, get_type(param.type))\n",
    "\n",
    "    def execute_try_stmt(self, stmt: TryStmt, environment: SymbolicEnvironment) -> SET:\n",
    "        tree = self.execute(stmt.block, environment)\n",
    "        result = (SETNode(environment, stmt.code), [tree])\n",
    "\n",
    "        for leaf in get_leaves(tree):\n",
    "            node, _ = leaf\n",
    "            exc = node.environment.abrupt_completion\n",
    "            if exc is None or isinstance(exc, AbruptCompletionNoException):\n",
    "                continue\n",
    "\n",
    "            if stmt.exc_type is not None:\n",
    "                caught_exc_type = getattr(sys.modules[\"builtins\"], stmt.exc_type)\n",
    "                if isinstance(exc, caught_exc_type):\n",
    "                    catch_block_tree = self.execute(stmt.except_block,\n",
    "                                                    node.environment.set_abrupt_completion(None))\n",
    "                    result = replace_in_tree(result, leaf, catch_block_tree)\n",
    "            else:\n",
    "                catch_block_tree = self.execute(stmt.block, node.environment.set_abrupt_completion(None))\n",
    "                result = replace_in_tree(result, leaf, catch_block_tree)\n",
    "\n",
    "        return result\n",
    "\n",
    "    def execute_if_stmt(self, stmt: IfStmt, environment: SymbolicEnvironment) -> SET:\n",
    "        eval_expr_result = self.evaluate(stmt.guard, environment)\n",
    "\n",
    "        result: SET = (SETNode(environment, stmt.code), [])\n",
    "\n",
    "        for value, constraint in eval_expr_result:\n",
    "            new_env = environment.add_constraint(constraint)\n",
    "            if self.unsatisfiable(new_env):\n",
    "                continue\n",
    "\n",
    "            if isinstance(value, Exception):\n",
    "                result[1].append((SETNode(new_env.set_abrupt_completion(value), ''), []))\n",
    "                continue\n",
    "\n",
    "            assert z3.is_bool(value)\n",
    "\n",
    "            then_constraint: z3.BoolRef = z3.simplify(value)\n",
    "            else_constraint = z3.simplify(z3.Not(value))\n",
    "\n",
    "            then_env = new_env.add_constraint(then_constraint)\n",
    "            else_env = new_env.add_constraint(else_constraint)\n",
    "\n",
    "            then_env_unsat = self.unsatisfiable(then_env)\n",
    "            else_env_unsat = self.unsatisfiable(else_env)\n",
    "\n",
    "            if not then_env_unsat:\n",
    "                subtree = self.execute(stmt.then_block, then_env)\n",
    "                result[1].append(subtree)\n",
    "\n",
    "            if not else_env_unsat:\n",
    "                if stmt.else_block is not None:\n",
    "                    subtree = self.execute(stmt.else_block, else_env)\n",
    "                    result[1].append(subtree)\n",
    "                else:\n",
    "                    result[1].append((SETNode(else_env, ''), []))\n",
    "\n",
    "        return result\n",
    "\n",
    "    def execute_while_stmt(self, stmt: WhileStmt, environment: SymbolicEnvironment) -> SET:\n",
    "        eval_guard_res = self.evaluate(stmt.guard, environment)\n",
    "        result = (SETNode(environment, stmt.code), [])\n",
    "\n",
    "        for value, constraint in eval_guard_res:\n",
    "            new_env = environment.add_constraint(constraint)\n",
    "            if self.unsatisfiable(new_env):\n",
    "                continue\n",
    "            \n",
    "            if isinstance(value, Exception):\n",
    "                exc_node = SETNode(new_env.set_abrupt_completion(value), '')\n",
    "                result[1].append((exc_node, []))\n",
    "                continue\n",
    "\n",
    "            then_constraint: z3.BoolRef = z3.simplify(value)\n",
    "            then_env = new_env.add_constraint(then_constraint)\n",
    "\n",
    "            if not self.unsatisfiable(then_env):\n",
    "                subtree = self.execute(stmt.body, then_env)\n",
    "                result[1].append(subtree)\n",
    "                for leaf in get_leaves(subtree):\n",
    "                    node, _ = leaf\n",
    "                    env = node.environment\n",
    "                    exc = env.abrupt_completion\n",
    "                    if exc is not None:\n",
    "                        if isinstance(exc, Continue):\n",
    "                            env = env.set_abrupt_completion(None)\n",
    "                        elif isinstance(exc, Break):\n",
    "                            env = env.set_abrupt_completion(None)\n",
    "                            result = replace_in_tree(\n",
    "                                result, leaf, (SETNode(env, ''), []))\n",
    "                            continue\n",
    "                        else:\n",
    "                            continue\n",
    "\n",
    "                    while_tree = self.execute(stmt, env)\n",
    "                    result = replace_in_tree(result, leaf, while_tree)\n",
    "                    continue\n",
    "\n",
    "            else_constraint: z3.BoolRef = z3.simplify(z3.Not(value))\n",
    "            else_env = new_env.add_constraint(else_constraint)\n",
    "\n",
    "            if not self.unsatisfiable(else_env):\n",
    "                if stmt.else_block is not None:\n",
    "                    subtree = self.execute(stmt.else_block, else_env)\n",
    "                    result[1].append((SETNode(else_env, 'else:'), [subtree]))\n",
    "                else:\n",
    "                    result[1].append((SETNode(else_env, ''), []))\n",
    "\n",
    "        return result\n",
    "\n",
    "    def execute_block(self, stmt: Block, environment: SymbolicEnvironment) -> SET:\n",
    "        return self.execute_sequence(tuple(stmt.stmts), environment)\n",
    "\n",
    "    def evaluate(self, expr: Expression, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        t = type(expr)\n",
    "        if t in self.expr_interpretations:\n",
    "            return [(value if isinstance(value, Exception) else z3.simplify(value), z3.simplify(constraint))\n",
    "                    for value, constraint in self.expr_interpretations[t](expr, environment)\n",
    "                    if not self.constraint_unsatisfiable(constraint)]\n",
    "\n",
    "        assert False\n",
    "\n",
    "    def evaluate_disjunction(self, expr: Disjunction, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        sub_results: List[EvalExprResult] = [self.evaluate(sub, environment) for sub in expr.conjunctions]\n",
    "\n",
    "        result: EvalExprResult = sub_results[0]\n",
    "        for sub_result in sub_results[1:]:\n",
    "            result = handle_binary_operation(result, sub_result, \"or\")\n",
    "\n",
    "        return result\n",
    "\n",
    "    def evaluate_conjunction(self, expr: Conjunction, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        sub_results: List[EvalExprResult] = [self.evaluate(sub, environment) for sub in expr.inversions]\n",
    "\n",
    "        result: EvalExprResult = sub_results[0]\n",
    "        for sub_result in sub_results[1:]:\n",
    "            result = handle_binary_operation(result, sub_result, \"and\")\n",
    "\n",
    "        return result\n",
    "\n",
    "    def evaluate_inversion(self, expr: Inversion, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        sub_result = self.evaluate(expr.inversion, environment)\n",
    "        return [(exc, constr) for exc, constr in sub_result if isinstance(exc, Exception)] + \\\n",
    "               [(z3.Not(value), constr) for value, constr in sub_result if not isinstance(value, Exception)]\n",
    "\n",
    "    def evaluate_comparison(self, expr: Comparison, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        left = self.evaluate(expr.left, environment)\n",
    "        right = self.evaluate(expr.right, environment)\n",
    "        return handle_binary_operation(left, right, expr.op)\n",
    "\n",
    "    def evaluate_sum(self, expr: Sum, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        left = self.evaluate(expr.left, environment)\n",
    "        right = self.evaluate(expr.right, environment)\n",
    "        return handle_binary_operation(left, right, expr.op)\n",
    "\n",
    "    def evaluate_term(self, expr: Term, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        left = self.evaluate(expr.left, environment)\n",
    "        right = self.evaluate(expr.right, environment)\n",
    "        return handle_binary_operation(left, right, expr.op)\n",
    "\n",
    "    def evaluate_factor(self, expr: Factor, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        eval_res = self.evaluate(expr.factor, environment)\n",
    "\n",
    "        if expr.symb == \"-\":\n",
    "            return [(result if isinstance(result, Exception) else -result, constr)\n",
    "                    for result, constr in eval_res]\n",
    "        else:\n",
    "            return eval_res\n",
    "\n",
    "    def evaluate_fun_call(self, expr: FunctionCall, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        result: EvalExprResult = []\n",
    "\n",
    "        f_name = expr.name\n",
    "        if f_name not in environment.functions:\n",
    "            raise NameError(f\"name '{f_name}' is not defined\")\n",
    "\n",
    "        args_result = [self.evaluate(arg, environment) for arg in expr.args]\n",
    "\n",
    "        combinations: List[Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]] = \\\n",
    "            list(itertools.product(*args_result))\n",
    "\n",
    "        combination: Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]\n",
    "        for combination in combinations:\n",
    "            values: List[SymbolicValueType] = []\n",
    "            path_constraint: z3.BoolRef = z3.BoolVal(True)\n",
    "            sub_result: Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]\n",
    "            for sub_value, sub_constraint in combination:\n",
    "                path_constraint = z3.simplify(z3.And(path_constraint, sub_constraint))\n",
    "                if isinstance(sub_value, Exception):\n",
    "                    result.append(wrap_exceptions((sub_value, path_constraint)))\n",
    "                    break\n",
    "\n",
    "                values.append(sub_value)\n",
    "\n",
    "            if len(values) != len(combination):\n",
    "                continue\n",
    "\n",
    "            if f_name in environment.functions:\n",
    "                params, ret_type, fun = environment.functions[f_name]\n",
    "                if len(values) != len(params):\n",
    "                    raise TypeError(f\"{f_name}({', '.join(map(str, params))}) takes {len(params)} \"\n",
    "                                    f\"argument(s) but {len(values)} were given\")\n",
    "\n",
    "                result.extend(fun(environment, *values))\n",
    "            else:\n",
    "                assert False\n",
    "\n",
    "        return result\n",
    "\n",
    "    def evaluate_tuple_access(self, expr: TupleAccess, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        target_res = self.evaluate(expr.atom, environment)\n",
    "        expr_res = self.evaluate(expr.expression, environment)\n",
    "        return self.handle_tuple_access(target_res, expr_res)\n",
    "\n",
    "    def handle_tuple_access(self, tuple_result: EvalExprResult, expr_result: EvalExprResult) -> EvalExprResult:\n",
    "        result: EvalExprResult = []\n",
    "\n",
    "        for atom_value, atom_constraint in tuple_result:\n",
    "            if isinstance(atom_value, Exception):\n",
    "                result.append((atom_value, atom_constraint))\n",
    "                continue\n",
    "\n",
    "            if not isinstance(atom_value.sort(), z3.SeqSortRef):\n",
    "                result.append((TypeError(f\"'{atom_value.sort()} is not subscriptable\"),\n",
    "                               atom_constraint))\n",
    "                continue\n",
    "\n",
    "            for expr_value, expr_constraint in expr_result:\n",
    "                conjunction = z3.And(atom_constraint, expr_constraint)\n",
    "                if isinstance(expr_value, Exception):\n",
    "                    result.append((expr_value, conjunction))\n",
    "                    continue\n",
    "\n",
    "                if expr_value.sort() != z3.IntSort():\n",
    "                    result.append((TypeError(f\"list indices must be integers, not {expr_value.sort()}\"),\n",
    "                                   conjunction))\n",
    "                    continue\n",
    "\n",
    "                atom_value: z3.SeqRef\n",
    "\n",
    "                # In Python, indices can be in the closed interval [-len ; len - 1]\n",
    "                length_constraint_1 = z3.And(z3.IntVal(0) <= expr_value,\n",
    "                                             expr_value < z3.Length(atom_value))  # [0 ; len - 1]\n",
    "                length_constraint_2 = z3.And(z3.IntVal(0) > expr_value,\n",
    "                                             expr_value >= z3.IntVal(-1) * z3.Length(atom_value))  # [-len ; -1]\n",
    "\n",
    "                expr = z3.simplify(atom_value[expr_value])\n",
    "                if expr.decl().kind() == z3.Z3_OP_ITE and \\\n",
    "                        implies(z3.And(conjunction, length_constraint_1), expr.children()[0]):\n",
    "                    expr = expr.children()[1]\n",
    "\n",
    "                result.append((expr, z3.And(conjunction, length_constraint_1)))\n",
    "\n",
    "                expr = z3.simplify(atom_value[z3.Length(atom_value) + expr_value])\n",
    "                if expr.decl().kind() == z3.Z3_OP_ITE and \\\n",
    "                        implies(z3.And(conjunction, length_constraint_2), expr.children()[0]):\n",
    "                    expr = expr.children()[1]\n",
    "\n",
    "                result.append((expr, z3.And(conjunction, length_constraint_2)))\n",
    "\n",
    "                result.append((IndexError(\"tuple index out of range\"),\n",
    "                               z3.And(conjunction, z3.Not(length_constraint_1), z3.Not(length_constraint_2))))\n",
    "\n",
    "        return result\n",
    "\n",
    "    def evaluate_name(self, expr: NameAtom, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        return [(environment[expr.name], z3.BoolVal(True))]\n",
    "\n",
    "    def evaluate_boolean(self, expr: BooleanAtom, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        return [(z3.BoolVal(expr.value), z3.BoolVal(True))]\n",
    "\n",
    "    def evaluate_int(self, expr: IntAtom, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        return [(z3.IntVal(expr.number), z3.BoolVal(True))]\n",
    "\n",
    "    def evaluate_tuple(self, expr: TupleNode, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        sub_results: List[EvalExprResult] = [self.evaluate(elem, environment) for elem in expr.elems]\n",
    "\n",
    "        if not sub_results:\n",
    "            return [(z3.Empty(z3.SeqSort(z3.IntSort())), z3.BoolVal(True))]\n",
    "        \n",
    "        result: OrderedSet[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]] = OrderedSet([])\n",
    "\n",
    "        combinations: List[Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]] = \\\n",
    "            list(itertools.product(*sub_results))\n",
    "\n",
    "        combination: Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]\n",
    "        for combination in combinations:\n",
    "            values: List[SymbolicValueType] = []\n",
    "            path_constraint: z3.BoolRef = z3.BoolVal(True)\n",
    "            sub_result: Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]\n",
    "            for sub_value, sub_constraint in combination:\n",
    "                path_constraint = z3.simplify(z3.And(path_constraint, sub_constraint))\n",
    "                if isinstance(sub_value, Exception):\n",
    "                    result.add(wrap_exceptions((sub_value, path_constraint)))\n",
    "                    break\n",
    "\n",
    "                values.append(sub_value)\n",
    "\n",
    "            if len(values) != len(combination):\n",
    "                continue\n",
    "\n",
    "            if len(values) == 1:\n",
    "                value = values[0]\n",
    "                result.add((value if isinstance(value.sort(), z3.SeqSortRef) else z3.Unit(value), path_constraint))\n",
    "            elif len(values) > 1:\n",
    "                result.add((z3.Concat(*[value if isinstance(value.sort(), z3.SeqSortRef) else z3.Unit(value)\n",
    "                                        for value in values]), path_constraint))\n",
    "\n",
    "        return [unwrap_exceptions(r) for r in result]\n",
    "\n",
    "\n",
    "class ExceptionWrapper(Exception):\n",
    "    \"\"\"Wrapper for exception objects with support for equality checks and hashing.\"\"\"\n",
    "\n",
    "    def __init__(self, exc: Exception):\n",
    "        self.exc = exc\n",
    "\n",
    "    def __hash__(self):\n",
    "        return hash((type(self.exc), self.exc.args))\n",
    "\n",
    "    def __eq__(self, other):\n",
    "        return type(self.exc) is type(other.exc) and self.exc.args == other.exc.args\n",
    "\n",
    "\n",
    "def wrap_exceptions(r: Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]) -> \\\n",
    "        Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]:\n",
    "    value, constraint = r\n",
    "    return ExceptionWrapper(value) if isinstance(value, Exception) else value, constraint\n",
    "\n",
    "\n",
    "def unwrap_exceptions(r: Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]) -> \\\n",
    "        Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]:\n",
    "    value, constraint = r\n",
    "    return value.exc if isinstance(value, ExceptionWrapper) else value, constraint\n",
    "\n",
    "\n",
    "def get_leaves(tree: SET) -> List[SET]:\n",
    "    node, children = tree\n",
    "    if not children:\n",
    "        return [tree]\n",
    "    else:\n",
    "        return [leaf for child in children for leaf in get_leaves(child)]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Let's give it a try:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/markdown": [
       "```python\n",
       "x = 2 * y\n",
       "if x < 0:\n",
       "    x = -x\n",
       "```"
      ],
      "text/plain": [
       "<IPython.core.display.Markdown object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "image/svg+xml": [
       "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n",
       "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"\n",
       " \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n",
       "<!-- Generated by graphviz version 2.43.0 (0)\n",
       " -->\n",
       "<!-- Title: %3 Pages: 1 -->\n",
       "<svg width=\"524pt\" height=\"308pt\"\n",
       " viewBox=\"0.00 0.00 523.50 308.00\" xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n",
       "<g id=\"graph0\" class=\"graph\" transform=\"scale(1 1) rotate(0) translate(4 304)\">\n",
       "<title>%3</title>\n",
       "<polygon fill=\"white\" stroke=\"transparent\" points=\"-4,4 -4,-304 519.5,-304 519.5,4 -4,4\"/>\n",
       "<!-- 0 -->\n",
       "<g id=\"node1\" class=\"node\">\n",
       "<title>0</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"358,-300 171,-300 171,-248 358,-248 358,-300\"/>\n",
       "<text text-anchor=\"start\" x=\"183.5\" y=\"-280.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {x: x, y: y})</text>\n",
       "<text text-anchor=\"start\" x=\"219.5\" y=\"-259.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">x = (2 * y)</text>\n",
       "</g>\n",
       "<!-- 1 -->\n",
       "<g id=\"node2\" class=\"node\">\n",
       "<title>1</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"363.5,-212 165.5,-212 165.5,-160 363.5,-160 363.5,-212\"/>\n",
       "<text text-anchor=\"start\" x=\"177.5\" y=\"-192.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {x: 2·y, y: y})</text>\n",
       "<text text-anchor=\"start\" x=\"219\" y=\"-171.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">if (x &lt; 0):</text>\n",
       "</g>\n",
       "<!-- 0&#45;&gt;1 -->\n",
       "<g id=\"edge1\" class=\"edge\">\n",
       "<title>0&#45;&gt;1</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M264.5,-247.66C264.5,-239.7 264.5,-230.76 264.5,-222.24\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"268,-222.16 264.5,-212.16 261,-222.16 268,-222.16\"/>\n",
       "</g>\n",
       "<!-- 2 -->\n",
       "<g id=\"node3\" class=\"node\">\n",
       "<title>2</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"261,-124 2,-124 2,-72 261,-72 261,-124\"/>\n",
       "<text text-anchor=\"start\" x=\"14.5\" y=\"-104.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(y ≥ 0)}, None, {x: 2·y, y: y})</text>\n",
       "<text text-anchor=\"start\" x=\"107\" y=\"-83.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">x = &#45;x</text>\n",
       "</g>\n",
       "<!-- 1&#45;&gt;2 -->\n",
       "<g id=\"edge2\" class=\"edge\">\n",
       "<title>1&#45;&gt;2</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M225.67,-159.89C211.19,-150.53 194.57,-139.78 179.37,-129.95\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"180.9,-126.78 170.61,-124.29 177.1,-132.66 180.9,-126.78\"/>\n",
       "</g>\n",
       "<!-- 4 -->\n",
       "<g id=\"node5\" class=\"node\">\n",
       "<title>4</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"515.5,-116 279.5,-116 279.5,-80 515.5,-80 515.5,-116\"/>\n",
       "<text text-anchor=\"start\" x=\"291.5\" y=\"-93.8\" font-family=\"Times,serif\" font-size=\"14.00\">({y ≥ 0}, None, {x: 2·y, y: y})</text>\n",
       "</g>\n",
       "<!-- 1&#45;&gt;4 -->\n",
       "<g id=\"edge4\" class=\"edge\">\n",
       "<title>1&#45;&gt;4</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M303.33,-159.89C322.04,-147.79 344.31,-133.39 362.5,-121.63\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"364.57,-124.46 371.06,-116.1 360.76,-118.59 364.57,-124.46\"/>\n",
       "</g>\n",
       "<!-- 3 -->\n",
       "<g id=\"node4\" class=\"node\">\n",
       "<title>3</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"263,-36 0,-36 0,0 263,0 263,-36\"/>\n",
       "<text text-anchor=\"start\" x=\"12.5\" y=\"-13.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(y ≥ 0)}, None, {x: &#45;2·y, y: y})</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;3 -->\n",
       "<g id=\"edge3\" class=\"edge\">\n",
       "<title>2&#45;&gt;3</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M131.5,-71.89C131.5,-63.84 131.5,-54.88 131.5,-46.66\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"135,-46.41 131.5,-36.41 128,-46.41 135,-46.41\"/>\n",
       "</g>\n",
       "</g>\n",
       "</svg>\n"
      ],
      "text/plain": [
       "<graphviz.dot.Digraph at 0x7fb03c1dff40>"
      ]
     },
     "execution_count": 16,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "program = \"\"\"\n",
    "x = 2 * y\n",
    "if x < 0:\n",
    "    x = -x\n",
    "\"\"\"\n",
    "\n",
    "display_program(program)\n",
    "\n",
    "x = Variable(\"x\", INT_TYPE)\n",
    "y = Variable(\"y\", INT_TYPE)\n",
    "\n",
    "env = SymbolicEnvironment(SymbolicStore({x: x.to_z3(), y: y.to_z3()}))\n",
    "interpreter = SymbolicInterpreter()\n",
    "tree = interpreter.execute(parse(program), env)\n",
    "\n",
    "display_set(tree)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "To try out something more complex, we revisit the linear search method from {ref}`minipy` (we slightly simplified the code, which originally should merely illustrate the extent of the minipy language)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/markdown": [
       "```python\n",
       "def find(needle: int, haystack: tuple) -> int:\n",
       "    i = 0\n",
       "    while i < len(haystack):\n",
       "        if haystack[i] == needle:\n",
       "            break\n",
       "        \n",
       "        i = i + 1\n",
       "    else:\n",
       "        return -1\n",
       "    \n",
       "    return i\n",
       "\n",
       "t = (1, 2, 3, 4, )\n",
       "r = find(x, t)\n",
       "```"
      ],
      "text/plain": [
       "<IPython.core.display.Markdown object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "program=\"\"\"\n",
    "def find(needle: int, haystack: tuple) -> int:\n",
    "    i = 0\n",
    "    while i < len(haystack):\n",
    "        if haystack[i] == needle:\n",
    "            break\n",
    "        \n",
    "        i = i + 1\n",
    "    else:\n",
    "        return -1\n",
    "    \n",
    "    return i\n",
    "\n",
    "t = (1, 2, 3, 4, )\n",
    "r = find(x, t)\n",
    "\"\"\"\n",
    "\n",
    "display_program(program)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Instead of executing this program with a concrete \"needle\" value, we use a symbolic input `x`. Let's have a look at the result:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "NameError: name 'len' is not defined (expected)\n"
     ]
    }
   ],
   "source": [
    "env = SymbolicEnvironment(SymbolicStore({x: x.to_z3()}))\n",
    "interpreter = SymbolicInterpreter()\n",
    "\n",
    "with ExpectError(print_traceback=False):\n",
    "    tree = interpreter.execute(parse(program), env)\n",
    "    display_set(tree)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Ooops... Apparently, the interpreter does not know the `len` function which is built into Python. This is not surprising: Even the concrete interpreter has to be told about existing built-in functions. For the symbolic interpreter, the situation is much more complex, since we cannot call a built-in Python function with symbolic arguments. There are two ways to approach this problem:\n",
    "\n",
    "1. Implement the `len` function in minipy and add it to our program.\n",
    "2. Summarize the `len` function with a symbolic expression.\n",
    "\n",
    "Option (2) corresponds to the concept of *function summaries* or *contracts* {cite}`ahrendt.beckert.ea-16` {cite}`baldoni.coppa.ea-18`. Such a summary is basically a conjunction of pre- and postcondition pairs, such that if a precondition holds in the current state, the postcondition can be assumed after function execution. We choose this option to teach the symbolic interpreter about the `len` function, since it would be rather impractical to add full implementations of all used library functions to each analyzed program. Furthermore, summaries are a central concept of *compositional symbolic execution* {cite}`anand.godefroid.ea-08` (see also {ref}`techniques:compositional_se`). It allows to modularly execute one piece of code at a time to derive or prove its summary, while invoking other functions by their summaries instead of executing them repeatedly.\n",
    "\n",
    "We extend the symbolic interpreter such that we can pass it a collection of predicates that will be considered in function calls whenever a function implementation is not available. Technically, we pass the interpreter a dictionary with predicate names as keys, and triples of argument sorts, target sort, and a Callable producing a symbolic value from the given arguments as values. For the `len` function:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "metadata": {},
   "outputs": [],
   "source": [
    "predicates = {\"len\": ((z3.SeqSort(z3.IntSort()),), z3.IntSort(), z3.Length)}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SymbolicInterpreter(SymbolicInterpreter):\n",
    "    def __init__(self,\n",
    "                 predicates: Optional[Dict[str, Tuple[Tuple[z3.SortRef, ...], z3.SortRef, Callable]]] = None):\n",
    "        super().__init__()\n",
    "        self.predicates: Dict[str, Tuple[Tuple[z3.SortRef], z3.SortRef, Callable]] = \\\n",
    "            {} if predicates is None else predicates\n",
    "\n",
    "    def evaluate_fun_call(self, expr: FunctionCall, environment: SymbolicEnvironment) -> EvalExprResult:\n",
    "        result: EvalExprResult = []\n",
    "\n",
    "        f_name = expr.name\n",
    "        if f_name not in environment.functions and f_name not in self.predicates:\n",
    "            raise NameError(f\"name '{f_name}' is not defined\")\n",
    "\n",
    "        args_result = [self.evaluate(arg, environment) for arg in expr.args]\n",
    "\n",
    "        combinations: List[Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]] = \\\n",
    "            list(itertools.product(*args_result))\n",
    "\n",
    "        combination: Tuple[Tuple[Union[SymbolicValueType, Exception], z3.BoolRef], ...]\n",
    "        for combination in combinations:\n",
    "            values: List[SymbolicValueType] = []\n",
    "            path_constraint: z3.BoolRef = z3.BoolVal(True)\n",
    "            sub_result: Tuple[Union[SymbolicValueType, Exception], z3.BoolRef]\n",
    "            for sub_value, sub_constraint in combination:\n",
    "                path_constraint = z3.simplify(z3.And(path_constraint, sub_constraint))\n",
    "                if isinstance(sub_value, Exception):\n",
    "                    result.append(wrap_exceptions((sub_value, path_constraint)))\n",
    "                    break\n",
    "\n",
    "                values.append(sub_value)\n",
    "\n",
    "            if len(values) != len(combination):\n",
    "                continue\n",
    "\n",
    "            if f_name in environment.functions:\n",
    "                params, ret_type, fun = environment.functions[f_name]\n",
    "                if len(values) != len(params):\n",
    "                    raise TypeError(f\"{f_name}({', '.join(map(str, params))}) takes {len(params)} \"\n",
    "                                    f\"argument(s) but {len(values)} were given\")\n",
    "\n",
    "                result.extend(fun(environment, *values))\n",
    "            elif f_name in self.predicates:\n",
    "                params_sorts, ret_sort, pred = self.predicates[f_name]\n",
    "                if len(values) != len(params_sorts):\n",
    "                    raise TypeError(f\"{f_name}({', '.join(map(str, params_sorts))}) takes {len(params_sorts)} \"\n",
    "                                    f\"argument(s) but {len(values)} were given\")\n",
    "                call_result: SymbolicValueType = pred(*values)\n",
    "                assert call_result.sort() == ret_sort\n",
    "                result.append((call_result, path_constraint))\n",
    "            else:\n",
    "                assert False\n",
    "\n",
    "        return result"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 21,
   "metadata": {
    "tags": [
     "output_scroll",
     "full-width"
    ]
   },
   "outputs": [
    {
     "data": {
      "image/svg+xml": [
       "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n",
       "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"\n",
       " \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n",
       "<!-- Generated by graphviz version 2.43.0 (0)\n",
       " -->\n",
       "<!-- Title: %3 Pages: 1 -->\n",
       "<svg width=\"3643pt\" height=\"308pt\"\n",
       " viewBox=\"0.00 0.00 3643.00 308.00\" xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n",
       "<g id=\"graph0\" class=\"graph\" transform=\"scale(1 1) rotate(0) translate(4 304)\">\n",
       "<title>%3</title>\n",
       "<polygon fill=\"white\" stroke=\"transparent\" points=\"-4,4 -4,-304 3639,-304 3639,4 -4,4\"/>\n",
       "<!-- 0 -->\n",
       "<g id=\"node1\" class=\"node\">\n",
       "<title>0</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1885.5,-300 1481.5,-300 1481.5,-248 1885.5,-248 1885.5,-300\"/>\n",
       "<text text-anchor=\"start\" x=\"1619.5\" y=\"-280.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {x: x})</text>\n",
       "<text text-anchor=\"start\" x=\"1493.5\" y=\"-259.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">def find(needle: int, haystack: tuple) &#45;&gt; int:</text>\n",
       "</g>\n",
       "<!-- 1 -->\n",
       "<g id=\"node2\" class=\"node\">\n",
       "<title>1</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1770,-212 1597,-212 1597,-160 1770,-160 1770,-212\"/>\n",
       "<text text-anchor=\"start\" x=\"1620\" y=\"-192.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {x: x})</text>\n",
       "<text text-anchor=\"start\" x=\"1609.5\" y=\"-171.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">t = (1, 2, 3, 4, )</text>\n",
       "</g>\n",
       "<!-- 0&#45;&gt;1 -->\n",
       "<g id=\"edge1\" class=\"edge\">\n",
       "<title>0&#45;&gt;1</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1683.5,-247.66C1683.5,-239.7 1683.5,-230.76 1683.5,-222.24\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1687,-222.16 1683.5,-212.16 1680,-222.16 1687,-222.16\"/>\n",
       "</g>\n",
       "<!-- 2 -->\n",
       "<g id=\"node3\" class=\"node\">\n",
       "<title>2</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1977,-124 1390,-124 1390,-72 1977,-72 1977,-124\"/>\n",
       "<text text-anchor=\"start\" x=\"1402.5\" y=\"-104.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4))))})</text>\n",
       "<text text-anchor=\"start\" x=\"1626\" y=\"-83.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">r = find(x, t)</text>\n",
       "</g>\n",
       "<!-- 1&#45;&gt;2 -->\n",
       "<g id=\"edge2\" class=\"edge\">\n",
       "<title>1&#45;&gt;2</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1683.5,-159.66C1683.5,-151.7 1683.5,-142.76 1683.5,-134.24\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1687,-134.16 1683.5,-124.16 1680,-134.16 1687,-134.16\"/>\n",
       "</g>\n",
       "<!-- 3 -->\n",
       "<g id=\"node4\" class=\"node\">\n",
       "<title>3</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"659,-36 0,-36 0,0 659,0 659,-36\"/>\n",
       "<text text-anchor=\"start\" x=\"12.5\" y=\"-13.8\" font-family=\"Times,serif\" font-size=\"14.00\">({1 = x}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: 0})</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;3 -->\n",
       "<g id=\"edge3\" class=\"edge\">\n",
       "<title>2&#45;&gt;3</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1389.7,-80.08C1166.59,-67.22 860.68,-49.6 634.79,-36.59\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"634.95,-33.09 624.77,-36.01 634.55,-40.08 634.95,-33.09\"/>\n",
       "</g>\n",
       "<!-- 4 -->\n",
       "<g id=\"node5\" class=\"node\">\n",
       "<title>4</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"1336,-36 677,-36 677,0 1336,0 1336,-36\"/>\n",
       "<text text-anchor=\"start\" x=\"689.5\" y=\"-13.8\" font-family=\"Times,serif\" font-size=\"14.00\">({2 = x}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: 1})</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;4 -->\n",
       "<g id=\"edge4\" class=\"edge\">\n",
       "<title>2&#45;&gt;4</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1466.5,-72C1368.18,-60.67 1254.09,-47.53 1164.13,-37.16\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1164.45,-33.68 1154.12,-36.01 1163.65,-40.63 1164.45,-33.68\"/>\n",
       "</g>\n",
       "<!-- 5 -->\n",
       "<g id=\"node6\" class=\"node\">\n",
       "<title>5</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"2013,-36 1354,-36 1354,0 2013,0 2013,-36\"/>\n",
       "<text text-anchor=\"start\" x=\"1366.5\" y=\"-13.8\" font-family=\"Times,serif\" font-size=\"14.00\">({3 = x}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: 2})</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;5 -->\n",
       "<g id=\"edge5\" class=\"edge\">\n",
       "<title>2&#45;&gt;5</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1683.5,-71.89C1683.5,-63.84 1683.5,-54.88 1683.5,-46.66\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1687,-46.41 1683.5,-36.41 1680,-46.41 1687,-46.41\"/>\n",
       "</g>\n",
       "<!-- 6 -->\n",
       "<g id=\"node7\" class=\"node\">\n",
       "<title>6</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"2690,-36 2031,-36 2031,0 2690,0 2690,-36\"/>\n",
       "<text text-anchor=\"start\" x=\"2043.5\" y=\"-13.8\" font-family=\"Times,serif\" font-size=\"14.00\">({4 = x}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: 3})</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;6 -->\n",
       "<g id=\"edge6\" class=\"edge\">\n",
       "<title>2&#45;&gt;6</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1900.5,-72C1998.82,-60.67 2112.91,-47.53 2202.87,-37.16\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"2203.35,-40.63 2212.88,-36.01 2202.55,-33.68 2203.35,-40.63\"/>\n",
       "</g>\n",
       "<!-- 7 -->\n",
       "<g id=\"node8\" class=\"node\">\n",
       "<title>7</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"3635,-36 2708,-36 2708,0 3635,0 3635,-36\"/>\n",
       "<text text-anchor=\"start\" x=\"2720.5\" y=\"-13.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(3 = x) ∧ ¬(4 = x) ∧ ¬(1 = x) ∧ ¬(2 = x)}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: &#45;1})</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;7 -->\n",
       "<g id=\"edge7\" class=\"edge\">\n",
       "<title>2&#45;&gt;7</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1977.04,-81.61C2224.39,-68.65 2578.28,-50.1 2836.64,-36.55\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"2837.14,-40.03 2846.94,-36.01 2836.77,-33.04 2837.14,-40.03\"/>\n",
       "</g>\n",
       "</g>\n",
       "</svg>\n"
      ],
      "text/plain": [
       "<graphviz.dot.Digraph at 0x7fb03c1dfca0>"
      ]
     },
     "execution_count": 21,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "env = SymbolicEnvironment(SymbolicStore({x: x.to_z3()}))\n",
    "interpreter = SymbolicInterpreter(predicates=predicates)\n",
    "\n",
    "tree = interpreter.execute(parse(program), env)\n",
    "display_set(tree)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Since the picture is a little oversized, here come the leaves of the tree:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "metadata": {},
   "outputs": [],
   "source": [
    "def display_leaves(tree: SET):\n",
    "    leaves = get_leaves(tree)\n",
    "    for leaf, _ in leaves:\n",
    "        display_html(leaf.environment.to_html())"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "({1&nbsp;=&nbsp;x}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: 0})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "({2&nbsp;=&nbsp;x}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: 1})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "({3&nbsp;=&nbsp;x}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: 2})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "({4&nbsp;=&nbsp;x}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: 3})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "({&not;(3&nbsp;=&nbsp;x)&nbsp;&and;&nbsp;&not;(4&nbsp;=&nbsp;x)&nbsp;&and;&nbsp;&not;(1&nbsp;=&nbsp;x)&nbsp;&and;&nbsp;&not;(2&nbsp;=&nbsp;x)}, None, {x: x, t: Concat(Unit(1), Concat(Unit(2), Concat(Unit(3), Unit(4)))), r: -1})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "display_leaves(tree)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The flat shape of the tree reveals one of the characteristics of our symbolic interpreter: Since function calls are evaluated as expressions, we do not see the SET corresponding to the execution of the function body. In {ref}`techniques:transparent_function_execution`, we describe how to transparently integrate SETs corresponding to called functions. Apart from that, the result is as expected: If `x` occurs in `t` (i.e., its value is between 1 and 4), the result value of variable `r` is `1 - x`, and -1 otherwise.\n",
    "\n",
    "But what happens if we also leave the value of `t` symbolic? In that case, the interpreter would continue unrolling the loop until the stack overflows (we do not demonstrate this here since a timeout is hard to implement in Python). In presence of a symbolic tuple `t`, we obtain an *unbounded loop*. A straightforward to deal with such unbounded loops solution is to only unroll loops until a maximum depth. We extend the symbolic interpreter with a depth bound for loops."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SymbolicInterpreter(SymbolicInterpreter):\n",
    "    def __init__(self,\n",
    "                 loop_unrolling_threshold: Optional[int] = None,\n",
    "                 predicates: Optional[Dict[str, Tuple[Tuple[z3.SortRef, ...], z3.SortRef, Callable]]] = None):\n",
    "        super().__init__(predicates)\n",
    "        self.loop_unrolling_threshold = loop_unrolling_threshold\n",
    "\n",
    "    def execute(self, stmt: ASTNode, environment: SymbolicEnvironment,\n",
    "                loop_unrolling_threshold: Optional[int] = None) -> SET:\n",
    "        self.logger.debug(\"Executing statement %s\", stmt.code)\n",
    "\n",
    "        loop_unrolling_threshold = (\n",
    "            loop_unrolling_threshold if loop_unrolling_threshold is not None\n",
    "            else self.loop_unrolling_threshold)\n",
    "\n",
    "        t = type(stmt)\n",
    "        if isinstance(stmt, WhileStmt):\n",
    "            return self.execute_while_stmt(stmt, environment, loop_unrolling_threshold)\n",
    "\n",
    "        if t in self.stmt_interpretations:\n",
    "            return self.stmt_interpretations[t](stmt, environment)\n",
    "\n",
    "        if isinstance(stmt, Expression):\n",
    "            # Expression Statements\n",
    "            return SETNode(environment, stmt.code), [(SETNode(environment.add_constraint(constraint), None), [])\n",
    "                                                     for _, constraint in self.evaluate(stmt, environment)]\n",
    "\n",
    "        assert False\n",
    "\n",
    "    def execute_while_stmt(self, stmt: WhileStmt, environment: SymbolicEnvironment,\n",
    "                           loop_unrolling_threshold: Optional[int] = None) -> SET:\n",
    "        loop_unrolling_threshold = -1 if loop_unrolling_threshold is None else loop_unrolling_threshold\n",
    "\n",
    "        eval_guard_res = self.evaluate(stmt.guard, environment)\n",
    "        result = (SETNode(environment, stmt.code), [])\n",
    "\n",
    "        for value, constraint in eval_guard_res:\n",
    "            new_env = environment.add_constraint(constraint)\n",
    "            if self.unsatisfiable(new_env):\n",
    "                continue\n",
    "\n",
    "            if isinstance(value, Exception):\n",
    "                exc_node = SETNode(new_env.set_abrupt_completion(value), '')\n",
    "                result[1].append((exc_node, []))\n",
    "                continue\n",
    "\n",
    "            then_constraint: z3.BoolRef = z3.simplify(value)\n",
    "            then_env = new_env.add_constraint(then_constraint)\n",
    "\n",
    "            if not self.unsatisfiable(then_env) and loop_unrolling_threshold != 0:\n",
    "                subtree = self.execute(stmt.body, then_env)\n",
    "                result[1].append(subtree)\n",
    "                for leaf in get_leaves(subtree):\n",
    "                    node, _ = leaf\n",
    "                    env = node.environment\n",
    "                    exc = env.abrupt_completion\n",
    "                    if exc is not None:\n",
    "                        if isinstance(exc, Continue):\n",
    "                            env = env.set_abrupt_completion(None)\n",
    "                        elif isinstance(exc, Break):\n",
    "                            env = env.set_abrupt_completion(None)\n",
    "                            result = replace_in_tree(\n",
    "                                result, leaf, (SETNode(env, ''), []))\n",
    "                            continue\n",
    "                        else:\n",
    "                            continue\n",
    "\n",
    "                    while_tree = self.execute(stmt, env, loop_unrolling_threshold - 1)\n",
    "                    result = replace_in_tree(result, leaf, while_tree)\n",
    "                    continue\n",
    "\n",
    "            else_constraint: z3.BoolRef = z3.simplify(z3.Not(value))\n",
    "            else_env = new_env.add_constraint(else_constraint)\n",
    "\n",
    "            if not self.unsatisfiable(else_env):\n",
    "                if stmt.else_block is not None:\n",
    "                    subtree = self.execute(stmt.else_block, else_env)\n",
    "                    result[1].append((SETNode(else_env, 'else:'), [subtree]))\n",
    "                else:\n",
    "                    result[1].append((SETNode(else_env, ''), []))\n",
    "\n",
    "        return result"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "This time, we directly execute the body of the `find` function to obtain an SET for the actual loop:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "metadata": {
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "text/markdown": [
       "```python\n",
       "i = 0\n",
       "while i < len(haystack):\n",
       "    if haystack[i] == needle:\n",
       "        break\n",
       "    i = i + 1\n",
       "else:\n",
       "    return -1\n",
       "\n",
       "return i\n",
       "```"
      ],
      "text/plain": [
       "<IPython.core.display.Markdown object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "program=\"\"\"i = 0\n",
    "while i < len(haystack):\n",
    "    if haystack[i] == needle:\n",
    "        break\n",
    "    i = i + 1\n",
    "else:\n",
    "    return -1\n",
    "\n",
    "return i\n",
    "\"\"\"\n",
    "\n",
    "display_program(program)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "metadata": {
    "tags": [
     "output_scroll",
     "full-width"
    ]
   },
   "outputs": [
    {
     "data": {
      "image/svg+xml": [
       "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n",
       "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"\n",
       " \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n",
       "<!-- Generated by graphviz version 2.43.0 (0)\n",
       " -->\n",
       "<!-- Title: %3 Pages: 1 -->\n",
       "<svg width=\"1703pt\" height=\"1504pt\"\n",
       " viewBox=\"0.00 0.00 1702.50 1504.00\" xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n",
       "<g id=\"graph0\" class=\"graph\" transform=\"scale(1 1) rotate(0) translate(4 1500)\">\n",
       "<title>%3</title>\n",
       "<polygon fill=\"white\" stroke=\"transparent\" points=\"-4,4 -4,-1500 1698.5,-1500 1698.5,4 -4,4\"/>\n",
       "<!-- 0 -->\n",
       "<g id=\"node1\" class=\"node\">\n",
       "<title>0</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1256,-1496 882,-1496 882,-1444 1256,-1444 1256,-1496\"/>\n",
       "<text text-anchor=\"start\" x=\"894\" y=\"-1476.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {needle: needle, haystack: haystack})</text>\n",
       "<text text-anchor=\"start\" x=\"1048\" y=\"-1455.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">i = 0</text>\n",
       "</g>\n",
       "<!-- 1 -->\n",
       "<g id=\"node2\" class=\"node\">\n",
       "<title>1</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1272,-1408 866,-1408 866,-1356 1272,-1356 1272,-1408\"/>\n",
       "<text text-anchor=\"start\" x=\"878\" y=\"-1388.8\" font-family=\"Times,serif\" font-size=\"14.00\">({}, None, {needle: needle, haystack: haystack, i: 0})</text>\n",
       "<text text-anchor=\"start\" x=\"961.5\" y=\"-1367.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">while (i &lt; len(haystack)):</text>\n",
       "</g>\n",
       "<!-- 0&#45;&gt;1 -->\n",
       "<g id=\"edge1\" class=\"edge\">\n",
       "<title>0&#45;&gt;1</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1069,-1443.66C1069,-1435.7 1069,-1426.76 1069,-1418.24\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1072.5,-1418.16 1069,-1408.16 1065.5,-1418.16 1072.5,-1418.16\"/>\n",
       "</g>\n",
       "<!-- 2 -->\n",
       "<g id=\"node3\" class=\"node\">\n",
       "<title>2</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1066,-1320 484,-1320 484,-1268 1066,-1268 1066,-1320\"/>\n",
       "<text text-anchor=\"start\" x=\"496\" y=\"-1300.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 0)}, None, {needle: needle, haystack: haystack, i: 0})</text>\n",
       "<text text-anchor=\"start\" x=\"663.5\" y=\"-1279.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">if (haystack[i] == needle):</text>\n",
       "</g>\n",
       "<!-- 1&#45;&gt;2 -->\n",
       "<g id=\"edge2\" class=\"edge\">\n",
       "<title>1&#45;&gt;2</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M983.17,-1355.89C947.78,-1345.54 906.62,-1333.5 870.3,-1322.88\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"871.26,-1319.51 860.68,-1320.06 869.29,-1326.23 871.26,-1319.51\"/>\n",
       "</g>\n",
       "<!-- 20 -->\n",
       "<g id=\"node21\" class=\"node\">\n",
       "<title>20</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1643.5,-1320 1084.5,-1320 1084.5,-1268 1643.5,-1268 1643.5,-1320\"/>\n",
       "<text text-anchor=\"start\" x=\"1097\" y=\"-1300.8\" font-family=\"Times,serif\" font-size=\"14.00\">({Length(haystack) ≤ 0}, None, {needle: needle, haystack: haystack, i: 0})</text>\n",
       "<text text-anchor=\"start\" x=\"1343.5\" y=\"-1279.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">else:</text>\n",
       "</g>\n",
       "<!-- 1&#45;&gt;20 -->\n",
       "<g id=\"edge20\" class=\"edge\">\n",
       "<title>1&#45;&gt;20</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1155.12,-1355.89C1190.64,-1345.54 1231.93,-1333.5 1268.38,-1322.88\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1269.41,-1326.22 1278.03,-1320.06 1267.45,-1319.5 1269.41,-1326.22\"/>\n",
       "</g>\n",
       "<!-- 3 -->\n",
       "<g id=\"node4\" class=\"node\">\n",
       "<title>3</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"522.5,-1232 83.5,-1232 83.5,-1135 522.5,-1135 522.5,-1232\"/>\n",
       "<text text-anchor=\"start\" x=\"96\" y=\"-1212.3\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 0), If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"96\" y=\"-1197.3\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"96\" y=\"-1182.3\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"96\" y=\"-1167.3\" font-family=\"Times,serif\" font-size=\"14.00\">needle}, None, {needle: needle, haystack: haystack, i: 0})</text>\n",
       "<text text-anchor=\"start\" x=\"282.5\" y=\"-1146.3\" font-family=\"Courier,monospace\" font-size=\"14.00\">break</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;3 -->\n",
       "<g id=\"edge3\" class=\"edge\">\n",
       "<title>2&#45;&gt;3</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M666.12,-1267.97C622.38,-1257.92 570.35,-1245.96 519.52,-1234.27\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"520.24,-1230.85 509.71,-1232.02 518.67,-1237.67 520.24,-1230.85\"/>\n",
       "</g>\n",
       "<!-- 6 -->\n",
       "<g id=\"node7\" class=\"node\">\n",
       "<title>6</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1005.5,-1232 552.5,-1232 552.5,-1135 1005.5,-1135 1005.5,-1232\"/>\n",
       "<text text-anchor=\"start\" x=\"565\" y=\"-1212.3\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 0), ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"565\" y=\"-1197.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"565\" y=\"-1182.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"565\" y=\"-1167.3\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 0})</text>\n",
       "<text text-anchor=\"start\" x=\"734\" y=\"-1146.3\" font-family=\"Courier,monospace\" font-size=\"14.00\">i = (i + 1)</text>\n",
       "</g>\n",
       "<!-- 2&#45;&gt;6 -->\n",
       "<g id=\"edge6\" class=\"edge\">\n",
       "<title>2&#45;&gt;6</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M775.93,-1267.84C776.21,-1260.1 776.54,-1251.25 776.87,-1242.27\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"780.37,-1242.25 777.24,-1232.13 773.38,-1241.99 780.37,-1242.25\"/>\n",
       "</g>\n",
       "<!-- 4 -->\n",
       "<g id=\"node5\" class=\"node\">\n",
       "<title>4</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"498.5,-1099 59.5,-1099 59.5,-1002 498.5,-1002 498.5,-1099\"/>\n",
       "<text text-anchor=\"start\" x=\"72\" y=\"-1079.3\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 0), If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"72\" y=\"-1064.3\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"72\" y=\"-1049.3\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"72\" y=\"-1034.3\" font-family=\"Times,serif\" font-size=\"14.00\">needle}, None, {needle: needle, haystack: haystack, i: 0})</text>\n",
       "<text text-anchor=\"start\" x=\"246\" y=\"-1013.3\" font-family=\"Courier,monospace\" font-size=\"14.00\">return i</text>\n",
       "</g>\n",
       "<!-- 3&#45;&gt;4 -->\n",
       "<g id=\"edge4\" class=\"edge\">\n",
       "<title>3&#45;&gt;4</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M294.27,-1134.84C292.75,-1126.53 291.15,-1117.81 289.58,-1109.26\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"293.01,-1108.55 287.77,-1099.35 286.13,-1109.82 293.01,-1108.55\"/>\n",
       "</g>\n",
       "<!-- 5 -->\n",
       "<g id=\"node6\" class=\"node\">\n",
       "<title>5</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"537.5,-955.5 2.5,-955.5 2.5,-879.5 537.5,-879.5 537.5,-955.5\"/>\n",
       "<text text-anchor=\"start\" x=\"15\" y=\"-936.3\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 0), If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"15\" y=\"-921.3\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"15\" y=\"-906.3\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"15\" y=\"-891.3\" font-family=\"Times,serif\" font-size=\"14.00\">needle}, SymbolicReturn(0), {needle: needle, haystack: haystack, i: 0})</text>\n",
       "</g>\n",
       "<!-- 4&#45;&gt;5 -->\n",
       "<g id=\"edge5\" class=\"edge\">\n",
       "<title>4&#45;&gt;5</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M275.73,-1001.84C274.92,-990.07 274.05,-977.47 273.24,-965.72\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"276.73,-965.39 272.55,-955.66 269.75,-965.87 276.73,-965.39\"/>\n",
       "</g>\n",
       "<!-- 7 -->\n",
       "<g id=\"node8\" class=\"node\">\n",
       "<title>7</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1008.5,-1099 555.5,-1099 555.5,-1002 1008.5,-1002 1008.5,-1099\"/>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-1079.3\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 0), ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-1064.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-1049.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-1034.3\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 1})</text>\n",
       "<text text-anchor=\"start\" x=\"675\" y=\"-1013.3\" font-family=\"Courier,monospace\" font-size=\"14.00\">while (i &lt; len(haystack)):</text>\n",
       "</g>\n",
       "<!-- 6&#45;&gt;7 -->\n",
       "<g id=\"edge7\" class=\"edge\">\n",
       "<title>6&#45;&gt;7</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M780.09,-1134.84C780.28,-1126.62 780.48,-1117.99 780.67,-1109.54\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"784.17,-1109.43 780.9,-1099.35 777.18,-1109.27 784.17,-1109.43\"/>\n",
       "</g>\n",
       "<!-- 8 -->\n",
       "<g id=\"node9\" class=\"node\">\n",
       "<title>8</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1008.5,-966 555.5,-966 555.5,-869 1008.5,-869 1008.5,-966\"/>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-946.3\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-931.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-916.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-901.3\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 1})</text>\n",
       "<text text-anchor=\"start\" x=\"671\" y=\"-880.3\" font-family=\"Courier,monospace\" font-size=\"14.00\">if (haystack[i] == needle):</text>\n",
       "</g>\n",
       "<!-- 7&#45;&gt;8 -->\n",
       "<g id=\"edge8\" class=\"edge\">\n",
       "<title>7&#45;&gt;8</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M782,-1001.84C782,-993.62 782,-984.99 782,-976.54\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"785.5,-976.35 782,-966.35 778.5,-976.35 785.5,-976.35\"/>\n",
       "</g>\n",
       "<!-- 17 -->\n",
       "<g id=\"node18\" class=\"node\">\n",
       "<title>17</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1604,-966 1026,-966 1026,-869 1604,-869 1604,-966\"/>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-946.3\" font-family=\"Times,serif\" font-size=\"14.00\">({Length(haystack) ≤ 1, ¬(Length(haystack) ≤ 0), ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-931.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-916.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-901.3\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 1})</text>\n",
       "<text text-anchor=\"start\" x=\"1294\" y=\"-880.3\" font-family=\"Courier,monospace\" font-size=\"14.00\">else:</text>\n",
       "</g>\n",
       "<!-- 7&#45;&gt;17 -->\n",
       "<g id=\"edge17\" class=\"edge\">\n",
       "<title>7&#45;&gt;17</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M975.56,-1001.93C1019.81,-991.05 1066.96,-979.46 1111.61,-968.49\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1112.52,-971.87 1121.4,-966.08 1110.85,-965.07 1112.52,-971.87\"/>\n",
       "</g>\n",
       "<!-- 9 -->\n",
       "<g id=\"node10\" class=\"node\">\n",
       "<title>9</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"537.5,-833 84.5,-833 84.5,-691 537.5,-691 537.5,-833\"/>\n",
       "<text text-anchor=\"start\" x=\"97\" y=\"-813.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), If(Length(haystack) ≤ 1,</text>\n",
       "<text text-anchor=\"start\" x=\"97\" y=\"-798.8\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_u(haystack, 1),</text>\n",
       "<text text-anchor=\"start\" x=\"97\" y=\"-783.8\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_i(haystack, 1)) =</text>\n",
       "<text text-anchor=\"start\" x=\"97\" y=\"-768.8\" font-family=\"Times,serif\" font-size=\"14.00\">needle, ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"97\" y=\"-753.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"97\" y=\"-738.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"97\" y=\"-723.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 1})</text>\n",
       "<text text-anchor=\"start\" x=\"290.5\" y=\"-702.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">break</text>\n",
       "</g>\n",
       "<!-- 8&#45;&gt;9 -->\n",
       "<g id=\"edge9\" class=\"edge\">\n",
       "<title>8&#45;&gt;9</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M636.2,-868.98C604.34,-858.6 569.91,-847.38 535.78,-836.26\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"536.75,-832.89 526.16,-833.12 534.59,-839.55 536.75,-832.89\"/>\n",
       "</g>\n",
       "<!-- 12 -->\n",
       "<g id=\"node13\" class=\"node\">\n",
       "<title>12</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1008.5,-833 555.5,-833 555.5,-691 1008.5,-691 1008.5,-833\"/>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-813.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), ¬(If(Length(haystack) ≤ 1,</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-798.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 1),</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-783.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 1)) =</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-768.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle), ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-753.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-738.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-723.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 1})</text>\n",
       "<text text-anchor=\"start\" x=\"737\" y=\"-702.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">i = (i + 1)</text>\n",
       "</g>\n",
       "<!-- 8&#45;&gt;12 -->\n",
       "<g id=\"edge12\" class=\"edge\">\n",
       "<title>8&#45;&gt;12</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M782,-868.77C782,-860.68 782,-852.07 782,-843.41\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"785.5,-843.24 782,-833.24 778.5,-843.24 785.5,-843.24\"/>\n",
       "</g>\n",
       "<!-- 10 -->\n",
       "<g id=\"node11\" class=\"node\">\n",
       "<title>10</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"525.5,-655 72.5,-655 72.5,-513 525.5,-513 525.5,-655\"/>\n",
       "<text text-anchor=\"start\" x=\"85\" y=\"-635.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), If(Length(haystack) ≤ 1,</text>\n",
       "<text text-anchor=\"start\" x=\"85\" y=\"-620.8\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_u(haystack, 1),</text>\n",
       "<text text-anchor=\"start\" x=\"85\" y=\"-605.8\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_i(haystack, 1)) =</text>\n",
       "<text text-anchor=\"start\" x=\"85\" y=\"-590.8\" font-family=\"Times,serif\" font-size=\"14.00\">needle, ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"85\" y=\"-575.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"85\" y=\"-560.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"85\" y=\"-545.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 1})</text>\n",
       "<text text-anchor=\"start\" x=\"266\" y=\"-524.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">return i</text>\n",
       "</g>\n",
       "<!-- 9&#45;&gt;10 -->\n",
       "<g id=\"edge10\" class=\"edge\">\n",
       "<title>9&#45;&gt;10</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M306.22,-690.93C305.64,-682.46 305.05,-673.75 304.46,-665.15\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"307.95,-664.85 303.78,-655.11 300.97,-665.33 307.95,-664.85\"/>\n",
       "</g>\n",
       "<!-- 11 -->\n",
       "<g id=\"node12\" class=\"node\">\n",
       "<title>11</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"550,-466.5 0,-466.5 0,-345.5 550,-345.5 550,-466.5\"/>\n",
       "<text text-anchor=\"start\" x=\"12\" y=\"-446.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), If(Length(haystack) ≤ 1,</text>\n",
       "<text text-anchor=\"start\" x=\"12\" y=\"-431.8\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_u(haystack, 1),</text>\n",
       "<text text-anchor=\"start\" x=\"12\" y=\"-416.8\" font-family=\"Times,serif\" font-size=\"14.00\">   seq.nth_i(haystack, 1)) =</text>\n",
       "<text text-anchor=\"start\" x=\"12\" y=\"-401.8\" font-family=\"Times,serif\" font-size=\"14.00\">needle, ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"12\" y=\"-386.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"12\" y=\"-371.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"12\" y=\"-356.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, SymbolicReturn(1), {needle: needle, haystack: haystack, i: 1})</text>\n",
       "</g>\n",
       "<!-- 10&#45;&gt;11 -->\n",
       "<g id=\"edge11\" class=\"edge\">\n",
       "<title>10&#45;&gt;11</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M289.45,-512.93C287.81,-500.95 286.11,-488.51 284.48,-476.54\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"287.95,-476.05 283.13,-466.62 281.01,-477 287.95,-476.05\"/>\n",
       "</g>\n",
       "<!-- 13 -->\n",
       "<g id=\"node14\" class=\"node\">\n",
       "<title>13</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1008.5,-655 555.5,-655 555.5,-513 1008.5,-513 1008.5,-655\"/>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-635.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), ¬(If(Length(haystack) ≤ 1,</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-620.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 1),</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-605.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 1)) =</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-590.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle), ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-575.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-560.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"568\" y=\"-545.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 2})</text>\n",
       "<text text-anchor=\"start\" x=\"675\" y=\"-524.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">while (i &lt; len(haystack)):</text>\n",
       "</g>\n",
       "<!-- 12&#45;&gt;13 -->\n",
       "<g id=\"edge13\" class=\"edge\">\n",
       "<title>12&#45;&gt;13</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M782,-690.93C782,-682.46 782,-673.75 782,-665.15\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"785.5,-665.11 782,-655.11 778.5,-665.11 785.5,-665.11\"/>\n",
       "</g>\n",
       "<!-- 14 -->\n",
       "<g id=\"node15\" class=\"node\">\n",
       "<title>14</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1021.5,-477 568.5,-477 568.5,-335 1021.5,-335 1021.5,-477\"/>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-457.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), ¬(If(Length(haystack) ≤ 1,</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-442.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 1),</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-427.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 1)) =</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-412.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle), Length(haystack) ≤ 2, ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-397.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-382.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-367.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 2})</text>\n",
       "<text text-anchor=\"start\" x=\"774.5\" y=\"-346.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">else:</text>\n",
       "</g>\n",
       "<!-- 13&#45;&gt;14 -->\n",
       "<g id=\"edge14\" class=\"edge\">\n",
       "<title>13&#45;&gt;14</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M787.18,-512.93C787.8,-504.46 788.44,-495.75 789.08,-487.15\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"792.58,-487.34 789.82,-477.11 785.59,-486.83 792.58,-487.34\"/>\n",
       "</g>\n",
       "<!-- 15 -->\n",
       "<g id=\"node16\" class=\"node\">\n",
       "<title>15</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1021.5,-299 568.5,-299 568.5,-157 1021.5,-157 1021.5,-299\"/>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-279.8\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), ¬(If(Length(haystack) ≤ 1,</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-264.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 1),</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-249.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 1)) =</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-234.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle), Length(haystack) ≤ 2, ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-219.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-204.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"581\" y=\"-189.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 2})</text>\n",
       "<text text-anchor=\"start\" x=\"758\" y=\"-168.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">return &#45;1</text>\n",
       "</g>\n",
       "<!-- 14&#45;&gt;15 -->\n",
       "<g id=\"edge15\" class=\"edge\">\n",
       "<title>14&#45;&gt;15</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M795,-334.93C795,-326.46 795,-317.75 795,-309.15\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"798.5,-309.11 795,-299.11 791.5,-309.11 798.5,-309.11\"/>\n",
       "</g>\n",
       "<!-- 16 -->\n",
       "<g id=\"node17\" class=\"node\">\n",
       "<title>16</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"1072,-121 518,-121 518,0 1072,0 1072,-121\"/>\n",
       "<text text-anchor=\"start\" x=\"530\" y=\"-101.3\" font-family=\"Times,serif\" font-size=\"14.00\">({¬(Length(haystack) ≤ 1), ¬(If(Length(haystack) ≤ 1,</text>\n",
       "<text text-anchor=\"start\" x=\"530\" y=\"-86.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 1),</text>\n",
       "<text text-anchor=\"start\" x=\"530\" y=\"-71.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 1)) =</text>\n",
       "<text text-anchor=\"start\" x=\"530\" y=\"-56.3\" font-family=\"Times,serif\" font-size=\"14.00\">  needle), Length(haystack) ≤ 2, ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"530\" y=\"-41.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"530\" y=\"-26.3\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"530\" y=\"-11.3\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, SymbolicReturn(&#45;1), {needle: needle, haystack: haystack, i: 2})</text>\n",
       "</g>\n",
       "<!-- 15&#45;&gt;16 -->\n",
       "<g id=\"edge16\" class=\"edge\">\n",
       "<title>15&#45;&gt;16</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M795,-156.81C795,-148.36 795,-139.72 795,-131.27\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"798.5,-131.08 795,-121.08 791.5,-131.08 798.5,-131.08\"/>\n",
       "</g>\n",
       "<!-- 18 -->\n",
       "<g id=\"node19\" class=\"node\">\n",
       "<title>18</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1604,-810.5 1026,-810.5 1026,-713.5 1604,-713.5 1604,-810.5\"/>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-790.8\" font-family=\"Times,serif\" font-size=\"14.00\">({Length(haystack) ≤ 1, ¬(Length(haystack) ≤ 0), ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-775.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-760.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-745.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, None, {needle: needle, haystack: haystack, i: 1})</text>\n",
       "<text text-anchor=\"start\" x=\"1277.5\" y=\"-724.8\" font-family=\"Courier,monospace\" font-size=\"14.00\">return &#45;1</text>\n",
       "</g>\n",
       "<!-- 17&#45;&gt;18 -->\n",
       "<g id=\"edge18\" class=\"edge\">\n",
       "<title>17&#45;&gt;18</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1315,-868.77C1315,-853.61 1315,-836.68 1315,-820.88\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1318.5,-820.7 1315,-810.7 1311.5,-820.7 1318.5,-820.7\"/>\n",
       "</g>\n",
       "<!-- 19 -->\n",
       "<g id=\"node20\" class=\"node\">\n",
       "<title>19</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"1604,-622 1026,-622 1026,-546 1604,-546 1604,-622\"/>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-602.8\" font-family=\"Times,serif\" font-size=\"14.00\">({Length(haystack) ≤ 1, ¬(Length(haystack) ≤ 0), ¬(If(Length(haystack) ≤ 0,</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-587.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_u(haystack, 0),</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-572.8\" font-family=\"Times,serif\" font-size=\"14.00\">     seq.nth_i(haystack, 0)) =</text>\n",
       "<text text-anchor=\"start\" x=\"1038\" y=\"-557.8\" font-family=\"Times,serif\" font-size=\"14.00\">  needle)}, SymbolicReturn(&#45;1), {needle: needle, haystack: haystack, i: 1})</text>\n",
       "</g>\n",
       "<!-- 18&#45;&gt;19 -->\n",
       "<g id=\"edge19\" class=\"edge\">\n",
       "<title>18&#45;&gt;19</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1315,-713.37C1315,-688.14 1315,-657.25 1315,-632.1\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1318.5,-632.06 1315,-622.06 1311.5,-632.06 1318.5,-632.06\"/>\n",
       "</g>\n",
       "<!-- 21 -->\n",
       "<g id=\"node22\" class=\"node\">\n",
       "<title>21</title>\n",
       "<polygon fill=\"none\" stroke=\"black\" points=\"1643.5,-1209.5 1084.5,-1209.5 1084.5,-1157.5 1643.5,-1157.5 1643.5,-1209.5\"/>\n",
       "<text text-anchor=\"start\" x=\"1097\" y=\"-1190.3\" font-family=\"Times,serif\" font-size=\"14.00\">({Length(haystack) ≤ 0}, None, {needle: needle, haystack: haystack, i: 0})</text>\n",
       "<text text-anchor=\"start\" x=\"1327\" y=\"-1169.3\" font-family=\"Courier,monospace\" font-size=\"14.00\">return &#45;1</text>\n",
       "</g>\n",
       "<!-- 20&#45;&gt;21 -->\n",
       "<g id=\"edge21\" class=\"edge\">\n",
       "<title>20&#45;&gt;21</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1364,-1267.84C1364,-1253.59 1364,-1235.55 1364,-1219.9\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1367.5,-1219.53 1364,-1209.53 1360.5,-1219.53 1367.5,-1219.53\"/>\n",
       "</g>\n",
       "<!-- 22 -->\n",
       "<g id=\"node23\" class=\"node\">\n",
       "<title>22</title>\n",
       "<polygon fill=\"#aeeeee\" stroke=\"black\" points=\"1694.5,-1068.5 1033.5,-1068.5 1033.5,-1032.5 1694.5,-1032.5 1694.5,-1068.5\"/>\n",
       "<text text-anchor=\"start\" x=\"1046\" y=\"-1046.3\" font-family=\"Times,serif\" font-size=\"14.00\">({Length(haystack) ≤ 0}, SymbolicReturn(&#45;1), {needle: needle, haystack: haystack, i: 0})</text>\n",
       "</g>\n",
       "<!-- 21&#45;&gt;22 -->\n",
       "<g id=\"edge22\" class=\"edge\">\n",
       "<title>21&#45;&gt;22</title>\n",
       "<path fill=\"none\" stroke=\"black\" d=\"M1364,-1157.24C1364,-1134.78 1364,-1102 1364,-1078.91\"/>\n",
       "<polygon fill=\"black\" stroke=\"black\" points=\"1367.5,-1078.74 1364,-1068.74 1360.5,-1078.74 1367.5,-1078.74\"/>\n",
       "</g>\n",
       "</g>\n",
       "</svg>\n"
      ],
      "text/plain": [
       "<graphviz.dot.Digraph at 0x7fb0541e81c0>"
      ]
     },
     "execution_count": 26,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "needle = Variable(\"needle\", INT_TYPE)\n",
    "haystack = Variable(\"haystack\", TUPLE_TYPE)\n",
    "\n",
    "env = SymbolicEnvironment(SymbolicStore({needle: needle.to_z3(), haystack: haystack.to_z3()}))\n",
    "interpreter = SymbolicInterpreter(predicates=predicates, loop_unrolling_threshold=2)\n",
    "tree = interpreter.execute(parse(program), env)\n",
    "\n",
    "display_set(tree)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "({&not;(Length(haystack)&nbsp;&le;&nbsp;0),&nbsp;If(Length(haystack)&nbsp;&le;&nbsp;0,<BR/>&nbsp;&nbsp;&nbsp;seq.nth_u(haystack,&nbsp;0),<BR/>&nbsp;&nbsp;&nbsp;seq.nth_i(haystack,&nbsp;0))&nbsp;=<BR/>needle}, SymbolicReturn(0), {needle: needle, haystack: haystack, i: 0})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "({&not;(Length(haystack)&nbsp;&le;&nbsp;1),&nbsp;If(Length(haystack)&nbsp;&le;&nbsp;1,<BR/>&nbsp;&nbsp;&nbsp;seq.nth_u(haystack,&nbsp;1),<BR/>&nbsp;&nbsp;&nbsp;seq.nth_i(haystack,&nbsp;1))&nbsp;=<BR/>needle,&nbsp;&not;(If(Length(haystack)&nbsp;&le;&nbsp;0,<BR/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;seq.nth_u(haystack,&nbsp;0),<BR/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;seq.nth_i(haystack,&nbsp;0))&nbsp;=<BR/>&nbsp;&nbsp;needle)}, SymbolicReturn(1), {needle: needle, haystack: haystack, i: 1})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "({&not;(Length(haystack)&nbsp;&le;&nbsp;1),&nbsp;&not;(If(Length(haystack)&nbsp;&le;&nbsp;1,<BR/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;seq.nth_u(haystack,&nbsp;1),<BR/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;seq.nth_i(haystack,&nbsp;1))&nbsp;=<BR/>&nbsp;&nbsp;needle),&nbsp;Length(haystack)&nbsp;&le;&nbsp;2,&nbsp;&not;(If(Length(haystack)&nbsp;&le;&nbsp;0,<BR/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;seq.nth_u(haystack,&nbsp;0),<BR/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;seq.nth_i(haystack,&nbsp;0))&nbsp;=<BR/>&nbsp;&nbsp;needle)}, SymbolicReturn(-1), {needle: needle, haystack: haystack, i: 2})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "({Length(haystack)&nbsp;&le;&nbsp;1,&nbsp;&not;(Length(haystack)&nbsp;&le;&nbsp;0),&nbsp;&not;(If(Length(haystack)&nbsp;&le;&nbsp;0,<BR/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;seq.nth_u(haystack,&nbsp;0),<BR/>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;seq.nth_i(haystack,&nbsp;0))&nbsp;=<BR/>&nbsp;&nbsp;needle)}, SymbolicReturn(-1), {needle: needle, haystack: haystack, i: 1})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "data": {
      "text/html": [
       "({Length(haystack)&nbsp;&le;&nbsp;0}, SymbolicReturn(-1), {needle: needle, haystack: haystack, i: 0})"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "display_leaves(tree)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "For each tuple index within our set bound of 2, we obtain two leaves: One where the searched-for element has been found (the abrupt completion flag is set to `SymbolicReturn(index)`, where `index` is the first position where `needle` has been found), and one where it has not been found (`SymbolicReturn(-1)`). Additionally, we have one leaf for the case where the tuple is empty (Length(haystack) ≤ 0).\n",
    "\n",
    "In {ref}`techniques:loops_static_se`, we introduce a technique based on *loop invariants* (summaries for loops), which allows to symbolically execute loops (in an online, static symbolic interpreter) without depth bounds.\n",
    "\n",
    "But what does it actually mean to stop SE after a given bound has exceeded? What is the effect of choosing coarse summaries for functions or loops? And how can we check whether the implementation of the symbolic interpreter discussed in this section is correct? These questions are related to the *semantics* of symbolic transition systems, which we discuss next."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## References\n",
    "\n",
    "```{bibliography}\n",
    ":filter: docname in docnames\n",
    "```\n",
    "\n",
    "[^floor-division]: See https://docs.python.org/3/reference/expressions.html#index-67"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.5"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
